import os
import json
import queue
import shlex
import pprint
import logging

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.cli as s_cli
import synapse.lib.cmd as s_cmd
import synapse.lib.node as s_node
import synapse.lib.time as s_time
import synapse.lib.msgpack as s_msgpack

logger = logging.getLogger(__name__)

RED = '#ff0066'
YELLOW = '#f4e842'
BLUE = '#6faef2'
DARKBLUE = '#4842f5'

PROVNEW_COLOR = DARKBLUE
UNKNOWN_COLOR = BLUE
NODEEDIT_COLOR = "lightblue"
WARNING_COLOR = YELLOW
SYNTAX_ERROR_COLOR = RED

class Log(s_cli.Cmd):
    '''Add a storm log to the local command session.

Notes:
    By default, the log file contains all messages received from the execution of
    a Storm query by the current CLI. By default, these messages are saved to a
    file located in ~/.syn/stormlogs/storm_(date).(format).

Examples:
    # Enable logging all messages to mpk files (default)
    log --on

    # Disable logging and close the current file
    log --off

    # Enable logging, but only log edits. Log them as jsonl instead of mpk.
    log --on --edits-only --format jsonl

    # Enable logging, but log to a custom path:
    log --on --path /my/aweome/log/directory/storm20010203.mpk

    # Log only the node messages which come back from a storm cmd execution.
    log --on --nodes-only --path /my/awesome/log/directory/stormnodes20010203.mpk
    '''
    _cmd_name = 'log'
    _cmd_syntax = (
        ('line', {'type': 'glob'}),  # type: ignore
    )

    splicetypes = (
        'tag:add',
        'tag:del',
        'node:add',
        'node:del',
        'prop:set',
        'prop:del',
        'tag:prop:set',
        'tag:prop:del',
    )

    def _make_argparser(self):

        parser = s_cmd.Parser(prog='log', outp=self, description=self.__doc__)
        muxp = parser.add_mutually_exclusive_group(required=True)
        muxp.add_argument('--on', action='store_true', default=False,
                          help='Enables logging of storm messages to a file.')
        muxp.add_argument('--off', action='store_true', default=False,
                          help='Disables message logging and closes the current storm file.')
        parser.add_argument('--format', choices=('mpk', 'jsonl'), default='mpk', type=str.lower,
                            help='The format used to save messages to disk. Defaults to msgpack (mpk).')
        parser.add_argument('--path', type=str, default=None,
                            help='The path to the log file.  This will append messages to a existing file.')
        optmux = parser.add_mutually_exclusive_group()
        optmux.add_argument('--edits-only', action='store_true', default=False,
                            help='Only records edits. Does not record any other messages.')
        optmux.add_argument('--nodes-only', action='store_true', default=False,
                            help='Only record the packed nodes returned by storm.')
        return parser

    def __init__(self, cli, **opts):
        s_cli.Cmd.__init__(self, cli, **opts)
        # Give ourselves a local ref to locs since we're stateful.
        self.locs = self._cmd_cli.locs
        self._cmd_cli.onfini(self.closeLogFd)

    def onStormMesg(self, mesg):
        self.locs.get('log:queue').put(mesg)

    @s_common.firethread
    def queueLoop(self):
        q = self.locs.get('log:queue')
        while not self._cmd_cli.isfini:
            try:
                mesg = q.get(timeout=2)
            except queue.Empty:
                continue
            smesg = mesg[1].get('mesg')
            self.save(smesg)

    def save(self, mesg):
        fd = self.locs.get('log:fd')
        editsonly = self.locs.get('log:editsonly')
        nodesonly = self.locs.get('log:nodesonly')
        if fd and not fd.closed:
            if editsonly and mesg[0] not in (*self.splicetypes, 'node:edits'):
                return
            if nodesonly:
                if mesg[0] != 'node':
                    return
                mesg = mesg[1]
            try:
                buf = self.encodeMsg(mesg)
            except Exception as e:  # pragma: no cover
                logger.error('Failed to serialize message: [%s]', str(e))
                return
            fd.write(buf)

    def encodeMsg(self, mesg):
        '''Get byts for a message'''

        fmt = self.locs.get('log:fmt')
        if fmt == 'jsonl':
            s = json.dumps(mesg, sort_keys=True) + '\n'
            buf = s.encode()
            return buf

        elif fmt == 'mpk':
            buf = s_msgpack.en(mesg)
            return buf

        mesg = f'Unknown encoding format: {fmt}'
        raise s_exc.SynErr(mesg=mesg)

    def closeLogFd(self):
        self._cmd_cli.off('storm:mesg', self.onStormMesg)
        q = self.locs.pop('log:queue', None)
        if q is not None:
            self.printf('Marking log queue done')
        thr = self.locs.pop('log:thr', None)
        if thr:
            self.printf('Joining log thread.')
            thr.join(2)
        fp = self.locs.pop('log:fp', None)
        fd = self.locs.pop('log:fd', None)
        for key in list(self.locs.keys()):
            if key.startswith('log:'):
                self.locs.pop(key, None)
        if fd:
            try:
                self.printf(f'Closing logfile: [{fp}]')
                fd.close()
            except Exception as e:  # pragma: no cover
                self.printf(f'Failed to close fd: [{str(e)}]')

    def openLogFd(self, opts):
        opath = self.locs.get('log:fp')
        if opath:
            self.printf('Must call --off to disable current file before starting a new file.')
            return
        fmt = opts.format
        path = opts.path
        nodes_only = opts.nodes_only
        edits_only = opts.edits_only
        if not path:
            ts = s_time.repr(s_common.now(), True)
            fn = f'storm_{ts}.{fmt}'
            path = s_common.getSynPath('stormlogs', fn)
        self.printf(f'Starting logfile at [{path}]')
        q = queue.Queue()
        fd = s_common.genfile(path)
        # Seek to the end of the file. Allows a user to append to a file.
        fd.seek(0, 2)
        self.locs['log:fp'] = path
        self.locs['log:fd'] = fd
        self.locs['log:fmt'] = fmt
        self.locs['log:queue'] = q
        self.locs['log:thr'] = self.queueLoop()
        self.locs['log:nodesonly'] = nodes_only
        self.locs['log:editsonly'] = edits_only
        self._cmd_cli.on('storm:mesg', self.onStormMesg)

    async def runCmdOpts(self, opts):

        line = opts.get('line', '')

        try:
            opts = self._make_argparser().parse_args(shlex.split(line))
        except s_exc.ParserExit:
            return

        if opts.on:
            return self.openLogFd(opts)

        if opts.off:
            return self.closeLogFd()

class StormCmd(s_cli.Cmd):
    '''
    Execute a storm query.

    Syntax:
        storm <query>

    Arguments:
        query: The storm query

    Optional Arguments:
        --hide-tags: Do not print tags.
        --hide-props: Do not print secondary properties.
        --hide-unknown: Do not print messages which do not have known handlers.
        --show-nodeedits:  Show full nodeedits (otherwise printed as a single . per edit).
        --editformat <format>: What format of edits the server shall emit.
                Options are
                   * nodeedits (default),
                   * splices (similar to < 2.0.0),
                   * count (just counts of nodeedits), or
                   * none (no such messages emitted).
        --show-prov:  Show provenance messages.
        --raw: Print the nodes in their raw format. This overrides --hide-tags and --hide-props.
        --debug: Display cmd debug information along with nodes in raw format. This overrides other display arguments.
        --path: Get path information about returned nodes.
        --show <names>: Limit storm events (server-side) to the comma-separated list.
        --file <path>: Run the storm query specified in the given file path.
        --optsfile <path>: Run the query with the given options from a JSON/YAML file.

    Examples:
        storm inet:ipv4=1.2.3.4
        storm --debug inet:ipv4=1.2.3.4

    '''
    _cmd_name = 'storm'

    editformat_enums = ('nodeedits', 'splices', 'count', 'none')
    _cmd_syntax = (
        ('--hide-tags', {}),  # type: ignore
        ('--show', {'type': 'valu'}),
        ('--show-nodeedits', {}),
        ('--show-prov', {}),
        ('--editformat', {'type': 'enum', 'defval': 'nodeedits', 'enum:vals': editformat_enums}),
        ('--file', {'type': 'valu'}),
        ('--optsfile', {'type': 'valu'}),
        ('--hide-props', {}),
        ('--hide-unknown', {}),
        ('--raw', {}),
        ('--debug', {}),
        ('--path', {}),
        ('--save-nodes', {'type': 'valu'}),
        ('query', {'type': 'glob'}),
    )

    def __init__(self, cli, **opts):
        s_cli.Cmd.__init__(self, cli, **opts)
        self.cmdmeths = {
            'node': self._onNode,
            'init': self._onInit,
            'fini': self._onFini,
            'print': self._onPrint,
            'warn': self._onWarn,
            'err': self._onErr,
            'node:edits': self._onNodeEdits,
            'node:edits:count': self._onNodeEditsCount,
            'prov:new': self._onProvNew,
        }
        self._indented = False

    def printf(self, mesg, addnl=True, color=None):
        if self._indented:
            s_cli.Cmd.printf(self, '')
            self._indented = False
        return s_cli.Cmd.printf(self, mesg, addnl=addnl, color=color)

    def _onNodeEdits(self, mesg, opts):
        edit = mesg[1]
        if not opts.get('show-nodeedits'):
            count = sum(len(e[2]) for e in edit.get('edits', ()))
            s_cli.Cmd.printf(self, '.' * count, addnl=False, color=NODEEDIT_COLOR)
            self._indented = True
            return

        self.printf(repr(mesg), color=NODEEDIT_COLOR)

    def _onNodeEditsCount(self, mesg, opts):
        count = mesg[1].get('count', 1)
        s_cli.Cmd.printf(self, '.' * count, addnl=False, color=NODEEDIT_COLOR)
        self._indented = True

    def _onProvNew(self, mesg, opts):
        if opts.get('show-prov'):
            self.printf(repr(mesg), color=PROVNEW_COLOR)

    def _printNodeProp(self, name, valu):
        self.printf(f'        {name} = {valu}')

    def _onNode(self, mesg, opts):
        node = mesg[1]

        if opts.get('raw'):
            self.printf(repr(node))
            return

        formname, formvalu = s_node.reprNdef(node)

        self.printf(f'{formname}={formvalu}')

        if not opts.get('hide-props'):

            for name in sorted(s_node.props(node).keys()):

                valu = s_node.reprProp(node, name)

                if name[0] != '.':
                    name = ':' + name
                self._printNodeProp(name, valu)

        if not opts.get('hide-tags'):

            for tag in sorted(s_node.tagsnice(node)):

                valu = s_node.reprTag(node, tag)
                tprops = s_node.reprTagProps(node, tag)
                printed = False
                if valu:
                    self.printf(f'        #{tag} = {valu}')
                    printed = True
                if tprops:
                    for prop, pval in tprops:
                        self.printf(f'        #{tag}:{prop} = {pval}')
                    printed = True
                if not printed:
                    self.printf(f'        #{tag}')

    def _onInit(self, mesg, opts):
        tick = mesg[1].get('tick')
        if tick is not None:
            tick = s_time.repr(tick)
            self.printf(f'Executing query at {tick}')

    def _onFini(self, mesg, opts):
        took = mesg[1].get('took')
        took = max(took, 1)

        count = mesg[1].get('count')
        pers = float(count) / float(took / 1000)
        self.printf('complete. %d nodes in %d ms (%d/sec).' % (count, took, pers))

    def _onPrint(self, mesg, opts):
        self.printf(mesg[1].get('mesg'))

    def _onWarn(self, mesg, opts):
        info = mesg[1]
        warn = info.pop('mesg', '')
        xtra = ', '.join([f'{k}={v}' for k, v in info.items()])
        if xtra:
            warn = ' '.join([warn, xtra])
        self.printf(f'WARNING: {warn}', color=WARNING_COLOR)

    def _onErr(self, mesg, opts):
        err = mesg[1]
        if err[0] == 'BadSyntax':
            pos = err[1].get('at', None)
            text = err[1].get('text', None)
            tlen = len(text)
            mesg = err[1].get('mesg', None)
            if pos is not None and text is not None and mesg is not None:
                text = text.replace('\n', ' ')
                # Handle too-long text
                if tlen > 60:
                    text = text[max(0, pos - 30):pos + 30]
                    if pos < tlen - 30:
                        text += '...'
                    if pos > 30:
                        text = '...' + text
                        pos = 33

                self.printf(text, color=BLUE)
                self.printf(f'{" "*pos}^', color=BLUE)
                self.printf(f'Syntax Error: {mesg}', color=SYNTAX_ERROR_COLOR)
                return

        self.printf(f'ERROR: {err}', color=SYNTAX_ERROR_COLOR)

    async def runCmdOpts(self, opts):

        text = opts.get('query')
        filename = opts.get('file')

        if bool(text) == bool(filename):
            self.printf('Cannot use a storm file and manual query together.')
            self.printf(self.__doc__)
            return

        if filename is not None:
            try:
                with open(filename, 'r') as fd:
                    text = fd.read()

            except FileNotFoundError:
                self.printf('file not found: %s' % (filename,))
                return

        stormopts = {}
        optsfile = opts.get('optsfile')
        if optsfile is not None:
            if not os.path.isfile(optsfile):
                self.printf('optsfile not found: %s' % (optsfile,))
                return
            stormopts = s_common.yamlload(optsfile)

        hide_unknown = opts.get('hide-unknown', self._cmd_cli.locs.get('storm:hide-unknown'))
        core = self.getCmdItem()

        stormopts.setdefault('repr', True)
        stormopts.setdefault('path', opts.get('path', False))

        showtext = opts.get('show')
        if showtext is not None:
            stormopts['show'] = showtext.split(',')

        editformat = opts['editformat']
        if editformat != 'nodeedits':
            stormopts['editformat'] = editformat

        nodesfd = None
        if opts.get('save-nodes'):
            nodesfd = s_common.genfile(opts.get('save-nodes'))
            nodesfd.truncate(0)

        try:

            async for mesg in core.storm(text, opts=stormopts):

                await self._cmd_cli.fire('storm:mesg', mesg=mesg)

                if opts.get('debug'):
                    self.printf(pprint.pformat(mesg))
                    continue

                if mesg[0] == 'node':

                    if nodesfd is not None:
                        byts = json.dumps(mesg[1]).encode()
                        nodesfd.write(byts + b'\n')

                try:
                    func = self.cmdmeths[mesg[0]]
                except KeyError:
                    if hide_unknown:
                        continue
                    self.printf(repr(mesg), color=UNKNOWN_COLOR)
                else:
                    func(mesg, opts)

        except s_exc.SynErr as e:

            if e.errinfo.get('errx') == 'CancelledError':
                self.printf('query canceled.')
                return

            raise

        finally:

            if nodesfd is not None:
                nodesfd.close()



# -*- coding: utf-8 -*-
# Stdlib
import os
import json
import logging

log = logging.getLogger(__name__)

ASSETS = os.path.split(__file__)[0]

def getAssetPath(fn):
    '''
    Get a asset file path.

    Args:
        fn (str): Asset to ask for.

    Returns:
        str: Path to asset file.

    Raises:
        ValueError: If path escaping detected or the path does not exist.
    '''
    fp = os.path.join(ASSETS, fn)
    absfp = os.path.abspath(fp)
    if not absfp.startswith(ASSETS):
        log.error('{} is not in {}'.format(fn, ASSETS))
        raise ValueError('Path escaping detected')
    if not os.path.isfile(absfp):
        log.error('{} does not exist'.format(absfp))
        raise ValueError('Asset does not exist')
    return absfp

def getAssetBytes(fn):
    '''
    Get the bytes for an assets file.

    Args:
        fn (str): File to get the bytes for.

    Returns:
        bytes: The bytes for the file.

    Raises:
        ValueError: If get_asset_path goes boom
    '''
    fp = getAssetPath(fn=fn)
    with open(fp, 'rb') as f:
        return f.read()

def getAssetStr(fn):
    '''
    Get the decoded string for an asset file.

    Args:
        fn (str): File to get and decode.

    Returns:
        str: The decoded string.
    '''
    byts = getAssetBytes(fn)
    return byts.decode()

def getAssets():
    '''
    Get a list of asset names.

    Returns:
        list: A list of file names in the assets.
    '''
    ret = []
    for fdir, fdirs, fns in os.walk(ASSETS):
        if '__pycache__' in fdir:
            continue
        for fn in fns:
            if fn in ['__init__.py']:
                continue
            fp = os.path.join(fdir, fn)
            relpath = os.path.relpath(fp, ASSETS)
            ret.append(relpath)
    return ret



import synapse.axon as s_axon
import synapse.cortex as s_cortex
import synapse.cryotank as s_cryotank

import synapse.lib.aha as s_aha
import synapse.lib.jsonstor as s_jsonstor

aha = s_aha.AhaCell
axon = s_axon.Axon
cortex = s_cortex.Cortex
cryotank = s_cryotank.CryoCell
jsonstor = s_jsonstor.JsonStorCell



import logging

import synapse.lib.module as s_module

logger = logging.getLogger(__name__)

ex_md5 = 'd41d8cd98f00b204e9800998ecf8427e'
ex_sha1 = 'da39a3ee5e6b4b0d3255bfef95601890afd80709'
ex_sha256 = 'ad9f4fe922b61e674a09530831759843b1880381de686a43460a76864ca0340c'
ex_sha384 = 'd425f1394e418ce01ed1579069a8bfaa1da8f32cf823982113ccbef531fa36bda9987f389c5af05b5e28035242efab6c'
ex_sha512 = 'ca74fe2ff2d03b29339ad7d08ba21d192077fece1715291c7b43c20c9136cd132788239189f3441a87eb23ce2660aa243f334295902c904b5520f6e80ab91f11'

x509vers = (
    (0, 'v1'),
    (2, 'v3'),
)

class CryptoModule(s_module.CoreModule):

    def getModelDefs(self):
        modl = {

            'types': (

                ('crypto:currency:transaction', ('guid', {}), {
                    'doc': 'An individual crypto currency transaction recorded on the blockchain.',
                }),
                ('crypto:currency:block', ('comp', {'fields': (
                                                        ('coin', 'crypto:currency:coin'),
                                                        ('offset', 'int'),
                                                   ), 'sepr': '/'}), {
                    'doc': 'An individual crypto currency block record on the blockchain.',
                }),
                ('crypto:smart:contract', ('guid', {}), {
                    'doc': 'A smart contract.',
                }),
                ('crypto:payment:input', ('guid', {}), {
                    'doc': 'A payment made into a transaction.',
                }),
                ('crypto:payment:output', ('guid', {}), {
                    'doc': 'A payment received from a transaction.',
                }),
                ('crypto:smart:token', ('comp', {'fields': (('contract', 'crypto:smart:contract'), ('tokenid', 'hugenum'))}), {
                    'doc': 'A token managed by a smart contract.',
                }),
                # TODO ('crypto:smart:method', ('guid', {}), {}),
                ('crypto:currency:coin', ('str', {'lower': True}), {
                    'doc': 'An individual crypto currency type.',
                    'ex': 'btc',
                }),
                ('crypto:currency:address', ('comp', {'fields': (('coin', 'crypto:currency:coin'), ('iden', 'str')), 'sepr': '/'}), {
                    'doc': 'An individual crypto currency address.',
                    'ex': 'btc/1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2',
                }),
                ('crypto:currency:client', ('comp', {'fields': (
                                                        ('inetaddr', 'inet:client'),
                                                        ('coinaddr', 'crypto:currency:address')
                                                    )}), {
                    'doc': 'A fused node representing a crypto currency address used by an Internet client.',
                    'ex': '(1.2.3.4, (btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2))',
                }),

                ('hash:md5', ('hex', {'size': 32}), {
                    'doc': 'A hex encoded MD5 hash.',
                    'ex': ex_md5
                }),
                ('hash:sha1', ('hex', {'size': 40}), {
                    'doc': 'A hex encoded SHA1 hash.',
                    'ex': ex_sha1
                }),
                ('hash:sha256', ('hex', {'size': 64}), {
                    'doc': 'A hex encoded SHA256 hash.',
                    'ex': ex_sha256
                }),
                ('hash:sha384', ('hex', {'size': 96}), {
                    'doc': 'A hex encoded SHA384 hash.',
                    'ex': ex_sha384
                }),
                ('hash:sha512', ('hex', {'size': 128}), {
                    'doc': 'A hex encoded SHA512 hash.',
                    'ex': ex_sha512
                }),
                ('hash:lm', ('hex', {'size': 32}), {
                    'doc': 'A hex encoded Microsoft Windows LM password hash.',
                    'ex': ex_md5
                }),
                ('hash:ntlm', ('hex', {'size': 32}), {
                    'doc': 'A hex encoded Microsoft Windows NTLM password hash.',
                    'ex': ex_md5
                }),

                ('rsa:key', ('comp', {'fields': (('mod', 'hex'), ('pub:exp', 'int')), }), {
                    'doc': 'An RSA keypair modulus and public exponent.'
                }),

                ('crypto:x509:cert', ('guid', {}), {
                    'doc': 'A unique X.509 certificate.',
                }),

                ('crypto:x509:san', ('comp', {'fields': (('type', 'str'), ('value', 'str'))}), {
                    'doc': 'An X.509 Subject Alternative Name (SAN).',
                }),

                ('crypto:x509:crl', ('guid', {}), {
                    'doc': 'A unique X.509 Certificate Revocation List.',
                }),

                ('crypto:x509:revoked', ('comp', {'fields': (('crl', 'crypto:x509:crl'), ('cert', 'crypto:x509:cert'))}), {
                    'doc': 'A revocation relationship between a CRL and an X.509 certificate.',
                }),

                ('crypto:x509:signedfile', ('comp', {'fields': (('cert', 'crypto:x509:cert'), ('file', 'file:bytes'))}), {
                    'doc': 'A digital signature relationship between an X.509 certificate and a file.',
                }),
            ),

            'forms': (

                ('crypto:payment:input', {}, (
                    ('transaction', ('crypto:currency:transaction', {}), {
                        'doc': 'The transaction the payment was input to.'}),
                    ('address', ('crypto:currency:address', {}), {
                        'doc': 'The address which paid into the transaction.'}),
                    ('value', ('econ:price', {}), {
                        'doc': 'The value of the currency paid into the transaction.'}),
                )),
                ('crypto:payment:output', {}, (
                    ('transaction', ('crypto:currency:transaction', {}), {
                        'doc': 'The transaction the payment was output from.'}),
                    ('address', ('crypto:currency:address', {}), {
                        'doc': 'The address which received payment from the transaction.'}),
                    ('value', ('econ:price', {}), {
                        'doc': 'The value of the currency recieved from the transaction.'}),
                )),
                ('crypto:currency:transaction', {}, (
                    ('hash', ('hex', {}), {
                        'doc': 'The unique transaction hash for the transaction.'}),
                    ('desc', ('str', {}), {
                        'doc': 'An analyst specified description of the transaction.'}),
                    ('block', ('crypto:currency:block', {}), {
                        'doc': 'The block which records the transaction.'}),
                    ('block:coin', ('crypto:currency:coin', {}), {
                        'doc': 'The coin/blockchain of the block which records this transaction.'}),
                    ('block:offset', ('int', {}), {
                        'doc': 'The offset of the block which records this transaction.'}),

                    ('success', ('bool', {}), {
                        'doc': 'Set to true if the transaction was successfully executed and recorded.'}),
                    ('status:code', ('int', {}), {
                        'doc': 'A coin specific status code which may represent an error reason.'}),
                    ('status:message', ('str', {}), {
                        'doc': 'A coin specific status message which may contain an error reason.'}),

                    ('to', ('crypto:currency:address', {}), {
                        'doc': 'The destination address of the transaction.'}),
                    ('from', ('crypto:currency:address', {}), {
                        'doc': 'The source address of the transaction.'}),
                    ('inputs', ('array', {'type': 'crypto:payment:input', 'sorted': True, 'uniq': True}), {
                        'deprecated': True,
                        'doc': 'Deprecated. Please use crypto:payment:input:transaction.'}),
                    ('outputs', ('array', {'type': 'crypto:payment:output', 'sorted': True, 'uniq': True}), {
                        'deprecated': True,
                        'doc': 'Deprecated. Please use crypto:payment:output:transaction.'}),
                    ('fee', ('econ:price', {}), {
                        'doc': 'The total fee paid to execute the transaction.'}),
                    ('value', ('econ:price', {}), {
                        'doc': 'The total value of the transaction.'}),
                    ('time', ('time', {}), {
                        'doc': 'The time this transaction was initiated.'}),

                    ('eth:gasused', ('int', {}), {
                        'doc': 'The amount of gas used to execute this transaction.'}),
                    ('eth:gaslimit', ('int', {}), {
                        'doc': 'The ETH gas limit specified for this transaction.'}),
                    ('eth:gasprice', ('econ:price', {}), {
                        'doc': 'The gas price (in ETH) specified for this transaction.'}),

                    ('contract:input', ('file:bytes', {}), {
                        'doc': 'Input value to a smart contract call.'}),
                    ('contract:output', ('file:bytes', {}), {
                        'doc': 'Output value of a smart contract call.'}),
                    # TODO break out args/retvals and maybe make humon repr?
                )),

                ('crypto:currency:block', {}, (
                    ('coin', ('crypto:currency:coin', {}), {
                        'doc': 'The coin/blockchain this block resides on.', 'ro': True, }),
                    ('offset', ('int', {}), {
                        'doc': 'The index of this block.', 'ro': True, }),
                    ('hash', ('hex', {}), {
                        'doc': 'The unique hash for the block.'}),
                    ('minedby', ('crypto:currency:address', {}), {
                        'doc': 'The address which mined the block.'}),
                    ('time', ('time', {}), {
                        'doc': 'Time timestamp embedded in the block by the miner.'}),
                )),

                ('crypto:smart:contract', {}, (
                    ('transaction', ('crypto:currency:transaction', {}), {
                        'doc': 'The transaction which created the contract.'}),
                    ('address', ('crypto:currency:address', {}), {
                        'doc': 'The address of the contract.'}),
                    ('bytecode', ('file:bytes', {}), {
                        'doc': 'The bytecode which implements the contract.'}),
                    ('token:name', ('str', {}), {
                        'doc': 'The ERC-20 token name.'}),
                    ('token:symbol', ('str', {}), {
                        'doc': 'The ERC-20 token symbol.'}),
                    ('token:totalsupply', ('hugenum', {}), {
                        'doc': 'The ERC-20 totalSupply value.'}),
                    # TODO methods, ABI conventions, source/disassembly
                )),
                ('crypto:smart:token', {}, (
                    ('contract', ('crypto:smart:contract', {}), {
                            'doc': 'The smart contract which defines and manages the token.', 'ro': True, }),
                    ('tokenid', ('hugenum', {}), {
                            'doc': 'The token ID.', 'ro': True, }),
                    ('owner', ('crypto:currency:address', {}), {
                            'doc': 'The address which currently owns the token.'}),
                    ('nft:url', ('inet:url', {}), {
                            'doc': 'The URL which hosts the NFT metadata.'}),
                    ('nft:meta', ('data', {}), {
                            'doc': 'The raw NFT metadata.'}),
                    ('nft:meta:name', ('str', {}), {
                            'doc': 'The name field from the NFT metadata.'}),
                    ('nft:meta:description', ('str', {}), {
                            'disp': {'hint': 'text'},
                            'doc': 'The description field from the NFT metadata.'}),
                    ('nft:meta:image', ('inet:url', {}), {
                            'doc': 'The image URL from the NFT metadata.'}),
                )),
                ('crypto:currency:coin', {}, (
                    ('name', ('str', {}), {
                        'doc': 'The full name of the crypto coin.'}),
                )),

                ('crypto:currency:address', {}, (
                    ('coin', ('crypto:currency:coin', {}), {
                        'doc': 'The crypto coin to which the address belongs.', 'ro': True, }),
                    ('iden', ('str', {}), {
                        'doc': 'The coin specific address identifier.', 'ro': True, }),
                    ('desc', ('str', {}), {
                        'doc': 'A free-form description of the address.'}),
                    ('contact', ('ps:contact', {}), {
                        'doc': 'Contact information associated with the address.'}),
                )),

                ('crypto:currency:client', {}, (
                    ('inetaddr', ('inet:client', {}), {
                        'doc': 'The Internet client address observed using the crypto currency address.', 'ro': True, }),
                    ('coinaddr', ('crypto:currency:address', {}), {
                        'doc': 'The crypto currency address observed in use the the Internet client.', 'ro': True, }),
                )),

                ('hash:md5', {}, ()),
                ('hash:sha1', {}, ()),
                ('hash:sha256', {}, ()),
                ('hash:sha384', {}, ()),
                ('hash:sha512', {}, ()),
                ('rsa:key', {}, (
                    ('mod', ('hex', {}), {'ro': True,
                       'doc': 'The RSA key modulus.'}),
                    ('pub:exp', ('int', {}), {'ro': True,
                       'doc': 'The public exponent of the key.'}),
                    ('bits', ('int', {}),
                     {'doc': 'The length of the modulus in bits.'}),
                    ('priv:exp', ('hex', {}),
                     {'doc': 'The private exponent of the key.'}),
                    ('priv:p', ('hex', {}),
                     {'doc': 'One of the two private primes.'}),
                    ('priv:q', ('hex', {}),
                     {'doc': 'One of the two private primes.'}),
                )),

                ('crypto:x509:signedfile', {}, (
                    ('cert', ('crypto:x509:cert', {}), {
                        'doc': 'The certificate for the key which signed the file.', 'ro': True, }),
                    ('file', ('file:bytes', {}), {
                        'doc': 'The file which was signed by the certificates key.', 'ro': True, }),
                )),

                ('crypto:x509:crl', {}, (
                    ('file', ('file:bytes', {}), {
                        'doc': 'The file containing the CRL.'}),
                    ('url', ('inet:url', {}), {
                        'doc': 'The URL where the CRL was published.'}),
                )),

                ('crypto:x509:revoked', {}, (
                    ('crl', ('crypto:x509:crl', {}), {
                        'doc': 'The CRL which revoked the certificate.', 'ro': True, }),
                    ('cert', ('crypto:x509:cert', {}), {
                        'doc': 'The certificate revoked by the CRL.', 'ro': True, }),
                )),

                ('crypto:x509:cert', {}, (

                    ('file', ('file:bytes', {}), {
                        'doc': 'The file that the certificate metadata was parsed from.',
                    }),

                    ('subject', ('str', {}), {
                        'doc': 'The subject identifier, commonly in X.500/LDAP format, to which the certificate was issued.',
                    }),

                    ('issuer', ('str', {}), {
                        'doc': 'The Distinguished Name (DN) of the Certificate Authority (CA) which issued the certificate.',
                    }),

                    ('issuer:cert', ('crypto:x509:cert', {}), {
                        'doc': 'The certificate used by the issuer to sign this certificate.',
                    }),

                    ('serial', ('str', {}), {
                        'doc': 'The serial number string in the certificate.',
                    }),

                    ('version', ('int', {'enums': x509vers}), {
                        'doc': 'The version integer in the certificate. (ex. 2 == v3 ).',
                    }),

                    ('validity:notbefore', ('time', {}), {
                        'doc': 'The timestamp for the beginning of the certificate validity period.',
                    }),

                    ('validity:notafter', ('time', {}), {
                        'doc': 'The timestamp for the end of the certificate validity period.',
                    }),

                    ('md5', ('hash:md5', {}), {
                        'doc': 'The MD5 fingerprint for the certificate.',
                    }),

                    ('sha1', ('hash:sha1', {}), {
                        'doc': 'The SHA1 fingerprint for the certificate.',
                    }),

                    ('sha256', ('hash:sha256', {}), {
                        'doc': 'The SHA256 fingerprint for the certificate.',
                    }),

                    ('rsa:key', ('rsa:key', {}), {
                        'doc': 'The optional RSA public key associated with the certificate.',
                    }),

                    ('algo', ('iso:oid', {}), {
                        'doc': 'The X.509 signature algorithm OID.',
                    }),

                    ('signature', ('hex', {}), {
                        'doc': 'The hexadecimal representation of the digital signature.',
                    }),

                    ('ext:sans', ('array', {'type': 'crypto:x509:san', 'uniq': True, 'sorted': True}), {
                        'doc': 'The Subject Alternate Names (SANs) listed in the certficate.',
                    }),

                    ('ext:crls', ('array', {'type': 'crypto:x509:san', 'uniq': True, 'sorted': True}), {
                        'doc': 'A list of Subject Alternate Names (SANs) for Distribution Points.',
                    }),

                    ('identities:fqdns', ('array', {'type': 'inet:fqdn', 'uniq': True, 'sorted': True}), {
                        'doc': 'The fused list of FQDNs identified by the cert CN and SANs.',
                    }),

                    ('identities:emails', ('array', {'type': 'inet:email', 'uniq': True, 'sorted': True}), {
                        'doc': 'The fused list of e-mail addresses identified by the cert CN and SANs.',
                    }),

                    ('identities:ipv4s', ('array', {'type': 'inet:ipv4', 'uniq': True, 'sorted': True}), {
                        'doc': 'The fused list of IPv4 addresses identified by the cert CN and SANs.',
                    }),

                    ('identities:ipv6s', ('array', {'type': 'inet:ipv6', 'uniq': True, 'sorted': True}), {
                        'doc': 'The fused list of IPv6 addresses identified by the cert CN and SANs.',
                    }),

                    ('identities:urls', ('array', {'type': 'inet:url', 'uniq': True, 'sorted': True}), {
                        'doc': 'The fused list of URLs identified by the cert CN and SANs.',
                    }),

                    ('crl:urls', ('array', {'type': 'inet:url', 'uniq': True, 'sorted': True}), {
                        'doc': 'The extracted URL values from the CRLs extension.',
                    }),

                    ('selfsigned', ('bool', {}), {
                        'doc': 'Whether this is a self-signed certificate.',
                    }),

                )),
            )
        }
        name = 'crypto'
        return ((name, modl),)



import pathlib

import synapse.exc as s_exc

import synapse.tests.utils as s_test
from synapse.tests.utils import alist

import synapse.lib.hive as s_hive
import synapse.lib.nexus as s_nexus
import synapse.lib.hiveauth as s_hiveauth

class AuthTest(s_test.SynTest):

    async def test_hive_auth(self):

        with self.getTestDir() as testdirn:

            async with self.getTestTeleHive() as hive:

                nexsroot = await s_nexus.NexsRoot.anit(testdirn)
                await nexsroot.startup(None)

                node = await hive.open(('hive', 'auth'))

                async with await s_hiveauth.Auth.anit(node, nexsroot=nexsroot) as auth:

                    auth.onfini(nexsroot.fini)

                    user = await auth.addUser('visi@vertex.link')
                    role = await auth.addRole('ninjas')

                    self.eq(user, auth.user(user.iden))
                    self.eq(user, await auth.getUserByName('visi@vertex.link'))

                    self.eq(role, auth.role(role.iden))
                    self.eq(role, await auth.getRoleByName('ninjas'))

                    with self.raises(s_exc.DupUserName):
                        await auth.addUser('visi@vertex.link')

                    with self.raises(s_exc.DupRoleName):
                        await auth.addRole('ninjas')

                    self.nn(user)

                    self.false(user.info.get('admin'))
                    self.len(0, user.info.get('rules'))
                    self.len(1, user.info.get('roles'))

                    await user.setAdmin(True)
                    self.true(user.info.get('admin'))
                    self.true(user.allowed(('foo', 'bar')))

                    await user.addRule((True, ('foo',)))
                    self.true(user.allowed(('foo', 'bar')))
                    self.len(1, user.permcache)

                    await user.delRule((True, ('foo',)))
                    self.len(0, user.permcache)

                    await user.addRule((True, ('foo',)))
                    await user.grant(role.iden)
                    self.len(0, user.permcache)
                    self.true(user.allowed(('baz', 'faz')))
                    self.len(1, user.permcache)

                    await role.addRule((True, ('baz', 'faz')))
                    self.len(0, user.permcache)
                    self.true(user.allowed(('baz', 'faz')))
                    self.len(1, user.permcache)

                    await user.setLocked(True)
                    self.false(user.allowed(('baz', 'faz')))

                    await user.setAdmin(False)
                    await user.setLocked(False)

                    self.true(user.allowed(('baz', 'faz')))
                    self.true(user.allowed(('foo', 'bar')))

                    # Add a DENY to the beginning of the rule list
                    await role.addRule((False, ('baz', 'faz')), indx=0)
                    self.false(user.allowed(('baz', 'faz')))

                    # Delete the DENY
                    await role.delRule((False, ('baz', 'faz')))

                    # After deleting, former ALLOW rule applies
                    self.true(user.allowed(('baz', 'faz')))

                    # non-existent rule returns default
                    self.none(user.allowed(('boo', 'foo')))
                    self.eq('yolo', user.allowed(('boo', 'foo'), default='yolo'))

                    await self.asyncraises(s_exc.NoSuchRole, user.revoke('newp'))

                    await user.revoke(role.iden)
                    self.none(user.allowed(('baz', 'faz')))

                    await user.grant(role.iden)
                    self.true(user.allowed(('baz', 'faz')))

                    await self.asyncraises(s_exc.NoSuchRole, auth.delRole('accountants'))

                    await auth.delRole(role.iden)
                    self.false(user.allowed(('baz', 'faz')))

                    await self.asyncraises(s_exc.NoSuchUser, auth.delUser('fred@accountancy.com'))

                    await auth.delUser(user.iden)
                    self.false(user.allowed(('baz', 'faz')))

                    role = await auth.addRole('lolusers')
                    role2 = await auth.addRole('lolusers2')

                    self.none(await role.setName('lolusers'))

                    with self.raises(s_exc.DupRoleName):
                        await role2.setName('lolusers')

                    await role.setName('roflusers')

                    self.nn(await auth.getRoleByName('roflusers'))
                    self.none(await auth.getRoleByName('lolusers'))

                    user = await auth.addUser('user1')
                    user2 = await auth.addUser('user')

                    # No problem if the user sets her own name to herself
                    self.none(await user.setName('user1'))

                    with self.raises(s_exc.DupUserName):
                        await user2.setName('user1')

                    await user.setName('user2')

                    self.nn(await auth.getUserByName('user2'))
                    self.none(await auth.getUserByName('user1'))

    async def test_hive_tele_auth(self):

        # confirm that the primitives used by higher level APIs
        # work using telepath remotes and property synchronize.

        async with self.getTestHiveDmon() as dmon:

            hive = dmon.shared.get('hive')

            hive.conf['auth:en'] = True

            auth = await hive.getHiveAuth()

            user = await auth.getUserByName('root')
            await user.setPasswd('secret')

            # tryPasswd
            self.true(await user.tryPasswd('secret'))
            self.false(await user.tryPasswd('beep'))
            self.false(await user.tryPasswd(None))

            # hive passwords must be non-zero length strings
            with self.raises(s_exc.BadArg):
                await user.setPasswd('')
            with self.raises(s_exc.BadArg):
                await user.setPasswd({'key': 'vau'})

            # passwords can be set to none, preventing tryPasswd from working
            await user.setPasswd(None)
            self.false(await user.tryPasswd(None))
            self.false(await user.tryPasswd('beep'))
            self.false(await user.tryPasswd('secret'))

            # Reset the password
            await user.setPasswd('secret')

            turl = self.getTestUrl(dmon, 'hive')

            # User can't access after being locked
            await user.setLocked(True)

            with self.raises(s_exc.AuthDeny):
                await s_hive.openurl(turl, user='root', passwd='secret')

            await user.setLocked(False)

            # User can't access after being unlocked with wrong password
            with self.raises(s_exc.AuthDeny):
                await s_hive.openurl(turl, user='root', passwd='newpnewp')

            # User can access with correct password after being unlocked with
            async with await s_hive.openurl(turl, user='root', passwd='secret'):
                await hive.open(('foo', 'bar'))

    async def test_hive_authgate_perms(self):

        async with self.getTestCoreAndProxy() as (core, prox):

            # We can retrieve the hivegate information
            gate = await prox.getAuthGate(core.view.iden)
            self.eq(gate['users'][0], {
                'iden': core.auth.rootuser.iden,
                'admin': True,
                'rules': (),
            })

            gates = await prox.getAuthGates()
            self.isin(core.view.iden, [g['iden'] for g in gates])

            fred = await prox.addUser('fred')
            bobo = await prox.addUser('bobo')
            await prox.setUserPasswd(fred['iden'], 'secret')
            await prox.setUserPasswd(bobo['iden'], 'secret')

            vdef2 = await core.view.fork()
            view2_iden = vdef2.get('iden')

            view2 = core.getView(view2_iden)

            await core.nodes('[test:int=10]')
            await view2.nodes('[test:int=11]')

            async with core.getLocalProxy(user='fred') as fredcore:
                viewopts = {'view': view2.iden}

                # Rando can access main view but not a fork
                self.eq(1, await fredcore.count('test:int'))

                await self.asyncraises(s_exc.AuthDeny, fredcore.count('test:int', opts=viewopts))

                viewiden = view2.iden
                layriden = view2.layers[0].iden

                # Add to a non-existent authgate
                rule = (True, ('view', 'read'))
                badiden = 'XXX'
                await self.asyncraises(s_exc.NoSuchAuthGate, prox.addUserRule(fred['iden'], rule, gateiden=badiden))

                # Rando can access forked view with explicit perms
                await prox.addUserRule(fred['iden'], rule, gateiden=viewiden)
                self.eq(2, await fredcore.count('test:int', opts=viewopts))

                friends = await prox.addRole('friends')

                # But still can't write to layer
                await self.asyncraises(s_exc.AuthDeny, fredcore.count('[test:int=12]', opts=viewopts))
                await self.asyncraises(s_exc.AuthDeny, fredcore.count('test:int=11 [:loc=us]', opts=viewopts))

                # fred can write to forked view's write layer with explicit perm through role

                rule = (True, ('node', 'prop', 'set',))
                await prox.addRoleRule(friends['iden'], rule, gateiden=layriden)

                # Before granting, still fails
                await self.asyncraises(s_exc.AuthDeny, fredcore.count('[test:int=12]', opts=viewopts))

                # After granting, succeeds
                await prox.addUserRole(fred['iden'], friends['iden'])
                self.eq(1, await fredcore.count('test:int=11 [:loc=ru]', opts=viewopts))

                # But adding a node still fails
                await self.asyncraises(s_exc.AuthDeny, fredcore.count('[test:int=12]', opts=viewopts))

                # After removing rule from friends, fails again
                await prox.delRoleRule(friends['iden'], rule, gateiden=layriden)
                await self.asyncraises(s_exc.AuthDeny, fredcore.count('test:int=11 [:loc=us]', opts=viewopts))

                rule = (True, ('node', 'add',))
                await prox.addUserRule(fred['iden'], rule, gateiden=layriden)
                self.eq(1, await fredcore.count('[test:int=12]', opts=viewopts))

                # Add an explicit DENY for adding test:int nodes
                rule = (False, ('node', 'add', 'test:int'))
                await prox.addUserRule(fred['iden'], rule, indx=0, gateiden=layriden)
                await self.asyncraises(s_exc.AuthDeny, fredcore.count('[test:int=13]', opts=viewopts))

                # Adding test:str is allowed though
                self.eq(1, await fredcore.count('[test:str=foo]', opts=viewopts))

                # An non-default world readable view works without explicit permission
                view2.worldreadable = True
                self.eq(3, await fredcore.count('test:int', opts=viewopts))

                # Deleting a user that has a role with an Authgate-specific rule
                rule = (True, ('node', 'prop', 'set',))
                await prox.addRoleRule(friends['iden'], rule, gateiden=layriden)
                self.eq(1, await fredcore.count('test:int=11 [:loc=sp]', opts=viewopts))
                await prox.addUserRole(bobo['iden'], friends['iden'])
                await prox.delAuthUser(bobo['iden'])
                self.eq(1, await fredcore.count('test:int=11 [:loc=us]', opts=viewopts))

                # Deleting a role removes all the authgate-specific role rules
                await prox.delRole(friends['iden'])
                await self.asyncraises(s_exc.AuthDeny, fredcore.count('test:int=11 [:loc=ru]', opts=viewopts))

                wlyr = view2.layers[0]

                await core.delView(view2.iden)
                await core.delLayer(wlyr.iden)

                # Verify that trashing the layer and view deletes the authgate from the hive
                self.none(core.auth.getAuthGate(wlyr.iden))
                self.none(core.auth.getAuthGate(view2.iden))

                # Verify that trashing the write layer deletes the remaining rules and backing store
                self.false(pathlib.Path(wlyr.dirn).exists())
                fred = await core.auth.getUserByName('fred')

                self.len(0, fred.getRules(gateiden=wlyr.iden))
                self.len(0, fred.getRules(gateiden=view2.iden))

    async def test_hive_auth_persistence(self):

        with self.getTestDir() as fdir:

            async with self.getTestCoreAndProxy(dirn=fdir) as (core, prox):

                # Set a bunch of permissions
                fred = await prox.addUser('fred')
                await prox.setUserPasswd(fred['iden'], 'secret')

                vdef2 = await core.view.fork()
                view2_iden = vdef2.get('iden')
                view2 = core.getView(view2_iden)

                await alist(core.eval('[test:int=10] [test:int=11]'))
                viewiden = view2.iden
                layriden = view2.layers[0].iden
                rule = (True, ('view', 'read',))
                await prox.addUserRule(fred['iden'], rule, gateiden=viewiden)
                friends = await prox.addAuthRole('friends')
                rule = (True, ('node', 'prop', 'set',))
                await prox.addRoleRule(friends['iden'], rule, gateiden=layriden)
                await prox.addUserRole(fred['iden'], friends['iden'])

            # Restart the core/auth and make sure perms work

            async with self.getTestCoreAndProxy(dirn=fdir) as (core, prox):
                async with core.getLocalProxy(user='fred') as fredcore:
                    viewopts = {'view': view2.iden}
                    self.eq(2, await fredcore.count('test:int', opts=viewopts))
                    self.eq(1, await fredcore.count('test:int=11 [:loc=ru]', opts=viewopts))

                await core.auth.delUser(fred['iden'])
                await core.auth.delRole(friends['iden'])

                self.none(await core.auth.getUserByName('fred'))
                self.none(await core.auth.getRoleByName('friends'))

            # restart after user/role removal and test they stayed gone
            async with self.getTestCoreAndProxy(dirn=fdir) as (core, prox):
                self.none(await core.auth.getUserByName('fred'))
                self.none(await core.auth.getRoleByName('friends'))

    async def test_hive_auth_invalid(self):

        async with self.getTestCore() as core:
            with self.raises(s_exc.BadArg):
                await core.auth.setRoleName(core.auth.allrole.iden, 'ninjas')
            with self.raises(s_exc.BadArg):
                await core.auth.rootuser.setName(1)
            with self.raises(s_exc.BadArg):
                await core.auth.allrole.setName(1)
            with self.raises(s_exc.SchemaViolation):
                await core.auth.rootuser.addRule('vi.si')
            with self.raises(s_exc.SchemaViolation):
                await core.auth.rootuser.setRules(None)
            with self.raises(s_exc.SchemaViolation):
                await core.auth.allrole.setRules(None)
            with self.raises(s_exc.BadArg):
                await core.auth.rootuser.setAdmin('lol')
            with self.raises(s_exc.BadArg):
                await core.auth.rootuser.setLocked('lol')
            with self.raises(s_exc.BadArg):
                await core.auth.rootuser.setArchived('lol')
            with self.raises(s_exc.SchemaViolation):
                await core.auth.allrole.addRule((1, ('hehe', 'haha')))
            with self.raises(s_exc.SchemaViolation):
                await core.auth.allrole.setRules([(True, ('hehe', 'haha'), 'newp')])
            with self.raises(s_exc.SchemaViolation):
                await core.auth.allrole.setRules([(True, )])



import os

import unittest.mock as mock

import synapse.common as s_common
import synapse.tests.utils as s_test

import synapse.lib.cell as s_cell
import synapse.lib.msgpack as s_msgpack

import synapse.tools.hive.load as s_hiveload

htree0 = {
    'kids': {
        'hehe': {'value': 20},
        'haha': {'value': 30, 'kids': {
            'foo': {'value': 'bar'},
            'baz': {'value': 'faz'},
        }},
    },
}

htree1 = {
    'kids': {
        'hehe': {'value': 20},
        'haha': {'value': 30, 'kids': {
            'baz': {'value': 'faz'},
        }},
    },
}

class HiveLoadTest(s_test.SynTest):

    async def test_tools_hiveload_vercheck(self):
        with self.getTestDir() as dirn:

            hivepath0 = os.path.join(dirn, 'hivesave0.mpk')
            s_msgpack.dumpfile(htree0, hivepath0)

            async with self.getTestHiveDmon() as dmon:
                hurl = self.getTestUrl(dmon, 'hive')

                argv = [hurl, hivepath0]

                def _getOldSynVers(self):
                    return (0, 0, 0)

                with mock.patch('synapse.telepath.Proxy._getSynVers', _getOldSynVers):
                    outp = self.getTestOutp()
                    retn = await s_hiveload.main(argv, outp=outp)
                    outp.expect('Hive version 0.0.0 is outside of the hive.load supported range')
                    self.eq(1, retn)

    async def test_tools_hiveload(self):

        with self.getTestDir() as dirn:

            hivepath0 = os.path.join(dirn, 'hivepath0.mpk')
            hivepath1 = os.path.join(dirn, 'hivepath1.mpk')
            yamlpath0 = os.path.join(dirn, 'hivepath0.yaml')

            s_msgpack.dumpfile(htree0, hivepath0)
            s_msgpack.dumpfile(htree1, hivepath1)
            s_common.yamlsave(htree0, yamlpath0)

            async with self.getTestHiveDmon() as dmon:

                hive = dmon.shared.get('hive')
                hurl = self.getTestUrl(dmon, 'hive')

                retn = await s_hiveload.main([hurl, hivepath0])
                self.eq(0, retn)

                self.eq(20, await hive.get(('hehe',)))
                self.eq(30, await hive.get(('haha',)))
                self.eq('bar', await hive.get(('haha', 'foo')))
                self.eq('faz', await hive.get(('haha', 'baz')))

                retn = await s_hiveload.main([hurl, hivepath1])
                self.eq(0, retn)

                self.eq(20, await hive.get(('hehe',)))
                self.eq(30, await hive.get(('haha',)))
                self.eq('bar', await hive.get(('haha', 'foo')))
                self.eq('faz', await hive.get(('haha', 'baz')))

                retn = await s_hiveload.main(['--trim', hurl, hivepath1])
                self.eq(0, retn)

                self.eq(20, await hive.get(('hehe',)))
                self.eq(30, await hive.get(('haha',)))
                self.eq('faz', await hive.get(('haha', 'baz')))

                self.none(await hive.get(('haha', 'foo')))

            async with self.getTestHiveDmon() as dmon:

                hive = dmon.shared.get('hive')
                hurl = self.getTestUrl(dmon, 'hive')

                await s_hiveload.main(['--path', 'v/i/s/i', '--yaml', hurl, yamlpath0])
                self.eq('bar', await hive.get(('v', 'i', 's', 'i', 'haha', 'foo')))

            path = os.path.join(dirn, 'cell')
            async with await s_cell.Cell.anit(path) as cell:
                curl = cell.getLocalUrl()
                await s_hiveload.main(['--path', 'gronk', curl, hivepath0])
                self.eq('bar', await cell.hive.get(('gronk', 'haha', 'foo')))
                await s_hiveload.main(['--trim', '--path', 'gronk', curl, hivepath1])
                self.none(await cell.hive.get(('gronk', 'haha', 'foo')))
                await s_hiveload.main(['--path', 'v/i/s/i', '--yaml', curl, yamlpath0])
                self.eq('bar', await cell.hive.get(('v', 'i', 's', 'i', 'haha', 'foo')))



NOMINAL = 'nominal'
DEGRADED = 'degraded'
FAILED = 'failed'
HEALTH_PRIORITY = {FAILED: 1, DEGRADED: 2, NOMINAL: 3}

class HealthCheck(object):
    def __init__(self, iden):
        self.iden = iden
        self._status = NOMINAL
        self.components = []

    def pack(self):
        return {'iden': self.iden,
                'components': self.components,
                'status': self._status,
                }

    def update(self, name, status, mesg='', data=None):
        '''
        Append a new component to the Healcheck object.

        Args:
            name (str): Name of the reported component.
            status (str): nomdinal/degraded/failed status code.
            mesg (str): Optional message about the component status.
            data (dict): Optional arbitrary dictionary of additional metadata about the component.

        Returns:
            None
        '''
        # Allow the component a shot at reporting a failure state
        status = status.lower()
        self.setStatus(status)
        if data is None:
            data = {}
        self.components.append({'status': status,
                                'name': name,
                                'mesg': mesg,
                                'data': data,
                                })

    def getStatus(self):
        return self._status

    def setStatus(self, valu):
        valu = valu.lower()
        new_priority = HEALTH_PRIORITY.get(valu)
        if new_priority is None:
            raise ValueError(f'Value {valu} must be in {list(HEALTH_PRIORITY)}')
        # Can only degrade status, not improve it.
        if new_priority >= HEALTH_PRIORITY.get(self._status):
            return
        self._status = valu



import heapq
import asyncio

import synapse.common as s_common

import synapse.lib.coro as s_coro
import synapse.lib.msgpack as s_msgpack

class SlabSeqn:
    '''
    An append optimized sequence of byte blobs.

    Args:
        lenv (lmdb.Environment): The LMDB Environment.
        name (str): The name of the sequence.
    '''
    def __init__(self, slab, name: str) -> None:

        self.slab = slab
        self.db = self.slab.initdb(name)

        self.indx = self.nextindx()
        self.addevents = []
        self.offsevents = []  # type: ignore # List[Tuple[int, int, asyncio.Event]] as a heap
        self._waitcounter = 0

        # NOTE: This is intended to be publicly accessible
        # and therefore must always represent the true size.
        self.size = self.stat()['entries']

    def _wake_waiters(self):

        for evnt in self.addevents:
            evnt.set()

        while self.offsevents and self.offsevents[0][0] < self.indx:
            _, _, evnt = heapq.heappop(self.offsevents)
            evnt.set()

    def pop(self, offs):
        '''
        Pop a single entry at the given offset.
        '''
        byts = self.slab.pop(s_common.int64en(offs), db=self.db)
        if byts is not None:
            self.size -= 1
            return (offs, s_msgpack.un(byts))

    async def cull(self, offs):
        '''
        Remove entries up to (and including) the given offset.
        '''
        for itemoffs, valu in self.iter(0):

            if itemoffs > offs:
                return

            if self.slab.delete(s_common.int64en(itemoffs), db=self.db):
                self.size -= 1
            await asyncio.sleep(0)

    def add(self, item, indx=None):
        '''
        Add a single item to the sequence.
        '''
        if indx is not None:
            if indx >= self.indx:
                self.slab.put(s_common.int64en(indx), s_msgpack.en(item), append=True, db=self.db)
                self.indx = indx + 1
                self.size += 1
                self._wake_waiters()
                return indx

            oldv = self.slab.replace(s_common.int64en(indx), s_msgpack.en(item), db=self.db)
            if oldv is None:
                self.size += 1
            return indx

        indx = self.indx
        retn = self.slab.put(s_common.int64en(indx), s_msgpack.en(item), append=True, db=self.db)
        assert retn, "Not adding the largest index"

        self.indx += 1
        self.size += 1

        self._wake_waiters()

        return indx

    def first(self):

        for lkey, lval in self.slab.scanByFull(db=self.db):
            return s_common.int64un(lkey), s_msgpack.un(lval)

        return None

    def last(self):

        last = self.slab.last(db=self.db)
        if last is None:
            return None

        lkey, lval = last

        indx = s_common.int64un(lkey)
        return indx, s_msgpack.un(lval)

    def stat(self):
        return self.slab.stat(db=self.db)

    def save(self, items):
        '''
        Save a series of items to a sequence.

        Args:
            items (tuple): The series of items to save into the sequence.

        Returns:
            The index of the first item
        '''
        rows = []
        indx = self.indx

        size = 0
        tick = s_common.now()

        for item in items:

            byts = s_msgpack.en(item)

            size += len(byts)

            lkey = s_common.int64en(indx)
            indx += 1

            rows.append((lkey, byts))

        retn = self.slab.putmulti(rows, append=True, db=self.db)
        took = s_common.now() - tick

        assert retn, "Not adding the largest indices"

        self.size += retn[1]

        origindx = self.indx
        self.indx = indx

        self._wake_waiters()

        return {'indx': indx, 'size': size, 'count': len(items), 'time': tick, 'took': took, 'orig': origindx}

    def index(self):
        '''
        Return the current index to be used
        '''
        return self.indx

    def nextindx(self):
        '''
        Determine the next insert offset according to storage.

        Returns:
            int: The next insert offset.
        '''
        byts = self.slab.lastkey(db=self.db)
        if byts is None:
            return 0

        return s_common.int64un(byts) + 1

    def iter(self, offs):
        '''
        Iterate over items in a sequence from a given offset.

        Args:
            offs (int): The offset to begin iterating from.

        Yields:
            (indx, valu): The index and valu of the item.
        '''
        startkey = s_common.int64en(offs)
        for lkey, lval in self.slab.scanByRange(startkey, db=self.db):
            offs = s_common.int64un(lkey)
            valu = s_msgpack.un(lval)
            yield offs, valu

    async def aiter(self, offs, wait=False, timeout=None):
        '''
        Iterate over items in a sequence from a given offset.

        Args:
            offs (int): The offset to begin iterating from.
            wait (boolean): Once caught up, yield new results in realtime.
            timeout (int): Max time to wait for a new item.

        Yields:
            (indx, valu): The index and valu of the item.
        '''
        startkey = s_common.int64en(offs)
        for lkey, lval in self.slab.scanByRange(startkey, db=self.db):
            offs = s_common.int64un(lkey)
            valu = s_msgpack.un(lval)
            yield offs, valu

        # no awaiting between here and evnt.timewait()
        if wait:
            evnt = s_coro.Event()
            try:
                self.addevents.append(evnt)
                while True:
                    evnt.clear()
                    if not await evnt.timewait(timeout=timeout):
                        return

                    startkey = s_common.int64en(offs + 1)
                    for lkey, lval in self.slab.scanByRange(startkey, db=self.db):
                        offs = s_common.int64un(lkey)
                        valu = s_msgpack.un(lval)
                        yield offs, valu
            finally:
                self.addevents.remove(evnt)

    async def gets(self, offs, wait=True):
        '''
        Returns an async generator of indx/valu tuples, optionally waiting and continuing to yield them as new entries
        are added

        Args:
            offs (int): The offset to begin iterating from.
            wait (bool):  Whether to continue yielding tupls when it hits the end of the sequence.

        Yields:
            (indx, valu): The index and valu of the item.
        '''
        while True:

            for (indx, valu) in self.iter(offs):
                yield (indx, valu)
                offs = indx + 1

            if not wait:
                return

            await self.waitForOffset(self.indx)

    def trim(self, offs):
        '''
        Delete entries starting at offset and moving forward.
        '''
        retn = False

        startkey = s_common.int64en(offs)
        for lkey, _ in self.slab.scanByRange(startkey, db=self.db):
            retn = True
            if self.slab.delete(lkey, db=self.db):
                self.size -= 1

        if retn:
            self.indx = self.nextindx()

        return retn

    def iterBack(self, offs):
        '''
        Iterate backwards over items in a sequence from a given offset.

        Args:
            offs (int): The offset to begin iterating from.

        Yields:
            (indx, valu): The index and valu of the item.
        '''
        startkey = s_common.int64en(offs)

        for lkey, lval in self.slab.scanByRangeBack(startkey, db=self.db):
            indx = s_common.int64un(lkey)
            valu = s_msgpack.un(lval)
            yield indx, valu

    def rows(self, offs):
        '''
        Iterate over raw indx, bytes tuples from a given offset.
        '''
        lkey = s_common.int64en(offs)
        for lkey, byts in self.slab.scanByRange(lkey, db=self.db):
            indx = s_common.int64un(lkey)
            yield indx, byts

    def get(self, offs):
        '''
        Retrieve a single row by offset
        '''
        lkey = s_common.int64en(offs)
        valu = self.slab.get(lkey, db=self.db)
        if valu is not None:
            return s_msgpack.un(valu)

    def getraw(self, byts):
        valu = self.slab.get(byts, db=self.db)
        if valu is not None:
            return s_msgpack.un(valu)

    def slice(self, offs, size):

        imax = size - 1

        for i, item in enumerate(self.iter(offs)):

            yield item

            if i == imax:
                break

    def sliceBack(self, offs, size):

        imax = size - 1

        for i, item in enumerate(self.iterBack(offs)):

            yield item

            if i == imax:
                break

    def getByIndxByts(self, indxbyts):
        byts = self.slab.get(indxbyts, db=self.db)
        if byts is not None:
            return s_msgpack.un(byts)

    def getOffsetEvent(self, offs):
        '''
        Returns an asyncio Event that will be set when the particular offset is written.  The event will be set if the
        offset has already been reached.
        '''
        evnt = asyncio.Event()

        if offs < self.indx:
            evnt.set()
            return evnt

        # We add a simple counter to the tuple to cause stable (and FIFO) sorting and prevent ties
        heapq.heappush(self.offsevents, (offs, self._waitcounter, evnt))

        self._waitcounter += 1

        return evnt

    async def waitForOffset(self, offs, timeout=None):
        '''
        Returns:
            true if the event got set, False if timed out
        '''

        if offs < self.indx:
            return True

        evnt = self.getOffsetEvent(offs)
        return await s_coro.event_wait(evnt, timeout=timeout)



import os
import subprocess

import yaml

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.tests.utils as s_t_utils

class CommonTest(s_t_utils.SynTest):
    def test_tuplify(self):
        tv = ['node', [['test:str', 'test'],
                       {'tags': {
                           'beep': [None, None],
                           'beep.boop': [0x01020304, 0x01020305]
                       },
                       'props': {
                           'tick': 0x01020304,
                           'tock': ['hehe', ['haha', 1]]
                       }}]]
        ev = ('node', (('test:str', 'test'),
                       {'tags': {
                           'beep': (None, None),
                           'beep.boop': (0x01020304, 0x01020305)
                       },
                       'props': {
                           'tick': 0x01020304,
                           'tock': ('hehe', ('haha', 1))
                       }}))

        self.eq(ev, s_common.tuplify(tv))
        tv = ('foo', ['bar', 'bat'])
        ev = ('foo', ('bar', 'bat'))
        self.eq(ev, s_common.tuplify(tv))

    def test_common_vertup(self):
        self.eq(s_common.vertup('1.3.30'), (1, 3, 30))
        self.true(s_common.vertup('30.40.50') > (9, 0))

    def test_common_file_helpers(self):
        # genfile
        with self.getTestDir() as testdir:
            fd = s_common.genfile(testdir, 'woot', 'foo.bin')
            fd.write(b'genfile_test')
            fd.close()

            with open(os.path.join(testdir, 'woot', 'foo.bin'), 'rb') as fd:
                buf = fd.read()
            self.eq(buf, b'genfile_test')

        # reqpath
        with self.getTestDir() as testdir:
            with s_common.genfile(testdir, 'test.txt') as fd:
                fd.write(b'')
            self.eq(os.path.join(testdir, 'test.txt'), s_common.reqpath(testdir, 'test.txt'))
            self.raises(s_exc.NoSuchFile, s_common.reqpath, testdir, 'newp')

        # reqfile
        with self.getTestDir() as testdir:
            with s_common.genfile(testdir, 'test.txt') as fd:
                fd.write(b'reqfile_test')
            fd = s_common.reqfile(testdir, 'test.txt')
            buf = fd.read()
            self.eq(buf, b'reqfile_test')
            fd.close()
            self.raises(s_exc.NoSuchFile, s_common.reqfile, testdir, 'newp')

        # getfile
        with self.getTestDir() as testdir:
            with s_common.genfile(testdir, 'test.txt') as fd:
                fd.write(b'getfile_test')
            fd = s_common.getfile(testdir, 'test.txt')
            buf = fd.read()
            self.eq(buf, b'getfile_test')
            fd.close()
            self.none(s_common.getfile(testdir, 'newp'))

        # getbytes
        with self.getTestDir() as testdir:
            with s_common.genfile(testdir, 'test.txt') as fd:
                fd.write(b'getbytes_test')
            buf = s_common.getbytes(testdir, 'test.txt')
            self.eq(buf, b'getbytes_test')
            self.none(s_common.getbytes(testdir, 'newp'))

        # reqbytes
        with self.getTestDir() as testdir:
            with s_common.genfile(testdir, 'test.txt') as fd:
                fd.write(b'reqbytes_test')
            buf = s_common.reqbytes(testdir, 'test.txt')
            self.eq(buf, b'reqbytes_test')
            self.raises(s_exc.NoSuchFile, s_common.reqbytes, testdir, 'newp')

        # listdir
        with self.getTestDir() as dirn:
            path = os.path.join(dirn, 'woot.txt')
            with open(path, 'wb') as fd:
                fd.write(b'woot')

            os.makedirs(os.path.join(dirn, 'nest'))
            with open(os.path.join(dirn, 'nest', 'nope.txt'), 'wb') as fd:
                fd.write(b'nope')

            retn = tuple(s_common.listdir(dirn))
            self.len(2, retn)

            retn = tuple(s_common.listdir(dirn, glob='*.txt'))
            self.eq(retn, ((path,)))

            # getDirSize: check against du
            real, appr = s_common.getDirSize(dirn)
            durealstr = subprocess.check_output(['du', '-B', '1', '-s', dirn])
            dureal = int(durealstr.split()[0])
            duapprstr = subprocess.check_output(['du', '-bs', dirn])
            duappr = int(duapprstr.split()[0])
            self.eq(dureal, real)
            self.eq(duappr, appr)

    def test_common_intify(self):
        self.eq(s_common.intify(20), 20)
        self.eq(s_common.intify("20"), 20)
        self.none(s_common.intify(None))
        self.none(s_common.intify("woot"))

    def test_common_guid(self):
        iden0 = s_common.guid()
        iden1 = s_common.guid('foo bar baz')
        iden2 = s_common.guid('foo bar baz')
        self.ne(iden0, iden1)
        self.eq(iden1, iden2)

    def test_common_isguid(self):
        self.true(s_common.isguid('98db59098e385f0bfdec8a6a0a6118b3'))
        self.false(s_common.isguid('visi'))

    def test_common_chunks(self):
        s = '123456789'
        parts = [chunk for chunk in s_common.chunks(s, 2)]
        self.eq(parts, ['12', '34', '56', '78', '9'])

        parts = [chunk for chunk in s_common.chunks(s, 100000)]
        self.eq(parts, [s])

        parts = [chunk for chunk in s_common.chunks(b'', 10000)]
        self.eq(parts, [b''])

        parts = [chunk for chunk in s_common.chunks([], 10000)]
        self.eq(parts, [[]])

        parts = [chunk for chunk in s_common.chunks('', 10000)]
        self.eq(parts, [''])

        parts = [chunk for chunk in s_common.chunks([1, 2, 3, 4, 5], 2)]
        self.eq(parts, [[1, 2], [3, 4], [5]])

        # set is unslicable
        with self.assertRaises(TypeError) as cm:
            parts = [chunk for chunk in s_common.chunks({1, 2, 3}, 10000)]

        # dict is unslicable
        with self.assertRaises(TypeError) as cm:
            parts = [chunk for chunk in s_common.chunks({1: 2}, 10000)]

        # empty dict is caught during the [0:0] slice
        with self.assertRaises(TypeError) as cm:
            parts = [chunk for chunk in s_common.chunks({}, 10000)]

    def test_common_lockfile(self):

        with self.getTestDir() as fdir:
            fp = os.path.join(fdir, 'hehe.lock')
            # Ensure that our yield is None
            with s_common.lockfile(fp) as cm:
                self.none(cm)

    def test_common_ehex_uhex(self):
        byts = b'deadb33f00010203'
        s = s_common.ehex(byts)
        self.isinstance(s, str)
        self.eq(s, '64656164623333663030303130323033')
        # uhex is a linear transform back
        obyts = s_common.uhex(s)
        self.isinstance(obyts, bytes)
        self.eq(byts, obyts)

    def test_common_buid(self):
        byts = b'deadb33f00010203'

        iden = s_common.buid()
        self.isinstance(iden, bytes)
        self.len(32, iden)
        # Buids are random by default
        iden2 = s_common.buid()
        self.ne(iden, iden2)

        # buids may be derived from any msgpackable valu which is stable
        iden3 = s_common.buid(byts)
        evalu = b'\xde\x8a\x8a\x88\xbc \xd4\xc1\x81J\xf5\xc7\xbf\xbc\xd2T6\xba\xd0\xf1\x10\xaa\x07<\xfa\xe5\xfc\x8c\x93\xeb\xb4 '
        self.len(32, iden3)
        self.eq(iden3, evalu)

    def test_common_spin(self):
        s = '1234'
        gen = iter(s)
        s_common.spin(gen)
        # Ensure we consumed everything from the generator
        self.raises(StopIteration, next, gen)

        # Consuming a generator could have effects!
        data = []
        def hehe():
            for c in s:
                data.append(c)
                yield c
        gen = hehe()
        s_common.spin(gen)
        self.eq(data, [c for c in s])

    def test_common_config(self):

        confdefs = (
            ('foo', {'defval': 20}),
            ('bar', {'defval': 30}),
        )

        conf = s_common.config({'foo': 80}, confdefs)

        self.eq(80, conf.get('foo'))
        self.eq(30, conf.get('bar'))

    def test_common_yaml(self):
        obj = [{'key': 1,
                'key2': [1, 2, 3],
                'key3': True,
                'key4': 'some str',
                'key5': {
                    'oh': 'my',
                    'we all': 'float down here'
                }, },
               'duck',
               False,
               'zero',
               0.1,
               ]
        with self.getTestDir() as dirn:
            s_common.yamlsave(obj, dirn, 'test.yaml')
            robj = s_common.yamlload(dirn, 'test.yaml')

            self.eq(obj, robj)

            obj = {'foo': 'bar', 'zap': [3, 4, 'f']}
            s_common.yamlsave(obj, dirn, 'test.yaml')
            s_common.yamlmod({'bar': 42}, dirn, 'test.yaml')
            robj = s_common.yamlload(dirn, 'test.yaml')
            obj['bar'] = 42
            self.eq(obj, robj)

            # Test yaml helper safety
            s = '!!python/object/apply:os.system ["pwd"]'
            with s_common.genfile(dirn, 'explode.yaml') as fd:
                fd.write(s.encode())
            self.raises(yaml.YAMLError, s_common.yamlload, dirn, 'explode.yaml')

    def test_switchext(self):
        retn = s_common.switchext('foo.txt', ext='.rdf')
        self.eq(retn, s_common.genpath('foo.rdf'))

        retn = s_common.switchext('.vim', ext='.rdf')
        self.eq(retn, s_common.genpath('.vim.rdf'))

    def test_envbool(self):
        with self.setTstEnvars(SYN_FOO='true', SYN_BAR='1', SYN_BAZ='foobar'):
            self.true(s_common.envbool('SYN_FOO'))
            self.true(s_common.envbool('SYN_BAR'))
            self.true(s_common.envbool('SYN_BAZ'))
            # non-existent and default behaviors
            self.false(s_common.envbool('SYN_NEWP'))
            self.true(s_common.envbool('SYN_NEWP', '1'))

        with self.setTstEnvars(SYN_FOO='false', SYN_BAR='0'):
            self.false(s_common.envbool('SYN_FOO'))
            self.false(s_common.envbool('SYN_BAR'))

    def test_normlog(self):
        self.eq(10, s_common.normLogLevel(' 10 '))
        self.eq(10, s_common.normLogLevel(10))
        self.eq(20, s_common.normLogLevel(' inFo\n'))
        with self.raises(s_exc.BadArg):
            s_common.normLogLevel(100)
        with self.raises(s_exc.BadArg):
            s_common.normLogLevel('BEEP')
        with self.raises(s_exc.BadArg):
            s_common.normLogLevel('12')
        with self.raises(s_exc.BadArg):
            s_common.normLogLevel({'key': 'newp'})

    async def test_merggenr(self):
        async def asyncl(data):
            for item in data:
                yield item

        async def alist(coro):
            return [x async for x in coro]

        l1 = (1, 2, 3)
        l2 = (4, 5, 6)
        l3 = (7, 8, 9)

        retn = s_common.merggenr([asyncl(lt) for lt in (l1, l2, l3)], lambda x, y: x < y)
        self.eq((1, 2, 3, 4, 5, 6, 7, 8, 9), await alist(retn))

        retn = s_common.merggenr([asyncl(lt) for lt in (l1, l1, l1)], lambda x, y: x < y)
        self.eq((1, 1, 1, 2, 2, 2, 3, 3, 3), await alist(retn))

        retn = s_common.merggenr([asyncl(lt) for lt in (l3, l2, l1)], lambda x, y: x < y)
        self.eq((1, 2, 3, 4, 5, 6, 7, 8, 9), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l2, l3, ())], lambda x: x)
        self.eq((1, 2, 3, 4, 5, 6, 7, 8, 9), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l1, l1)], lambda x: x)
        self.eq((1, 1, 1, 2, 2, 2, 3, 3, 3), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l3, l2, l1)], lambda x: x)
        self.eq((1, 2, 3, 4, 5, 6, 7, 8, 9), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l2, l3, ())])
        self.eq((1, 2, 3, 4, 5, 6, 7, 8, 9), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l1, l1)])
        self.eq((1, 1, 1, 2, 2, 2, 3, 3, 3), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l3, l2, l1)])
        self.eq((1, 2, 3, 4, 5, 6, 7, 8, 9), await alist(retn))

        l1 = (3, 2, 1)
        l2 = (6, 5, 4)
        l3 = (9, 8, 7)

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l2, l3)], lambda x: x, True)
        self.eq((9, 8, 7, 6, 5, 4, 3, 2, 1), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l1, l1)], lambda x: x, True)
        self.eq((3, 3, 3, 2, 2, 2, 1, 1, 1), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l3, l2, l1)], lambda x: x, True)
        self.eq((9, 8, 7, 6, 5, 4, 3, 2, 1), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l2, l3)], reverse=True)
        self.eq((9, 8, 7, 6, 5, 4, 3, 2, 1), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l1, l1, l1)], reverse=True)
        self.eq((3, 3, 3, 2, 2, 2, 1, 1, 1), await alist(retn))

        retn = s_common.merggenr2([asyncl(lt) for lt in (l3, l2, l1)], reverse=True)
        self.eq((9, 8, 7, 6, 5, 4, 3, 2, 1), await alist(retn))



import synapse.exc as s_exc
import synapse.common as s_common

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist

class TelcoModelTest(s_t_utils.SynTest):
    async def test_telco_simple(self):
        async with self.getTestCore() as core:

            typ = core.model.type('tel:mob:mcc')
            self.eq(typ.norm('001')[0], '001')
            self.raises(s_exc.BadTypeValu, typ.norm, '01')
            self.raises(s_exc.BadTypeValu, typ.norm, '0001')

            typ = core.model.type('tel:mob:mnc')
            self.eq(typ.norm('01')[0], '01')
            self.eq(typ.norm('001')[0], '001')
            self.raises(s_exc.BadTypeValu, typ.norm, '0001')
            self.raises(s_exc.BadTypeValu, typ.norm, '1')

            async with await core.snap() as snap:
                # tel:mob:tac
                oguid = s_common.guid()
                place = s_common.guid()
                props = {'manu': 'Acme Corp',
                         'model': 'eYephone 9000',
                         'internal': 'spYphone 9000',
                         'org': oguid,
                         }
                node = await snap.addNode('tel:mob:tac', 1, props)
                self.eq(node.ndef[1], 1)
                self.eq(node.get('manu'), 'acme corp')
                self.eq(node.get('model'), 'eyephone 9000')
                self.eq(node.get('internal'), 'spyphone 9000')
                self.eq(node.get('org'), oguid)
                # defvals
                node = await snap.addNode('tel:mob:tac', 2)

                # tel:mob:imid
                node = await snap.addNode('tel:mob:imid', (490154203237518, 310150123456789))
                self.eq(node.ndef[1], (490154203237518, 310150123456789))
                self.eq(node.get('imei'), 490154203237518)
                self.eq(node.get('imsi'), 310150123456789)

                # tel:mob:imsiphone
                node = await snap.addNode('tel:mob:imsiphone', (310150123456789, '+7(495) 124-59-83'))
                self.eq(node.ndef[1], (310150123456789, '74951245983'))
                self.eq(node.get('imsi'), 310150123456789)
                self.eq(node.get('phone'), '74951245983')

                # tel:mob:mcc
                node = await snap.addNode('tel:mob:mcc', '611', {'loc': 'gn'})
                self.eq(node.ndef[1], '611')
                self.eq(node.get('loc'), 'gn')

                # tel:mob:carrier
                node = await snap.addNode('tel:mob:carrier', ('001', '02'), {'org': oguid, 'loc': 'us'})
                self.eq(node.ndef[1], ('001', '02'))
                self.eq(node.get('mcc'), '001')
                self.eq(node.get('mnc'), '02')
                self.eq(node.get('org'), oguid)
                self.eq(node.get('loc'), 'us')

                # tel:mob:cell
                node = await snap.addNode('tel:mob:cell', (('001', '02'), 3, 4), {'radio': 'Pirate  ',
                                                                                  'latlong': (0, 0),
                                                                                  'place': place,
                                                                                  'loc': 'us.ca.la'})
                self.eq(node.get('carrier'), ('001', '02'))
                self.eq(node.get('carrier:mcc'), '001')
                self.eq(node.get('carrier:mnc'), '02')
                self.eq(node.get('lac'), 3)
                self.eq(node.get('cid'), 4)
                self.eq(node.get('loc'), 'us.ca.la')
                self.eq(node.get('radio'), 'pirate')
                self.eq(node.get('latlong'), (0.0, 0.0))
                self.eq(node.get('place'), place)

                await self.checkNodes(core, (('tel:mob:mcc', '001'),))

                # tel:mob:telem
                guid = s_common.guid()
                host = s_common.guid()
                softguid = s_common.guid()
                props = {'time': '2001',
                         'latlong': (-1, 1),
                         'place': place,
                         'host': host,
                         'loc': 'us',
                         'accuracy': '100mm',
                         'cell': (('001', '02'), 3, 4),
                         'imsi': '310150123456789',
                         'imei': '490154203237518',
                         'phone': '123 456 7890',
                         'mac': '00:00:00:00:00:00',
                         'ipv4': '1.2.3.4',
                         'ipv6': '::1',
                         'wifi': ('The Best SSID2', '00:11:22:33:44:55'),
                         'adid': 'someadid',
                         'aaid': 'somestr',
                         'idfa': 'someotherstr',
                         'name': 'Robert Grey',
                         'email': 'clown@vertex.link',
                         'acct': ('vertex.link', 'clown'),
                         'app': softguid,
                         'data': {'some key': 'some valu',
                                  'BEEP': 1}
                         }
                node = await snap.addNode('tel:mob:telem', guid, props)
                self.eq(node.ndef[1], guid)
                self.eq(node.get('time'), 978307200000)
                self.eq(node.get('latlong'), (-1.0, 1.0))
                self.eq(node.get('place'), place)
                self.eq(node.get('host'), host)
                self.eq(node.get('loc'), 'us')
                self.eq(node.get('accuracy'), 100)
                self.eq(node.get('cell'), (('001', '02'), 3, 4))
                self.eq(node.get('cell:carrier'), ('001', '02'))
                self.eq(node.get('imsi'), 310150123456789)
                self.eq(node.get('imei'), 490154203237518)
                self.eq(node.get('phone'), '1234567890')
                self.eq(node.get('mac'), '00:00:00:00:00:00')
                self.eq(node.get('ipv4'), 0x01020304)
                self.eq(node.get('ipv6'), '::1')
                self.eq(node.get('wifi'), ('The Best SSID2', '00:11:22:33:44:55')),
                self.eq(node.get('wifi:ssid'), 'The Best SSID2')
                self.eq(node.get('wifi:bssid'), '00:11:22:33:44:55')
                self.eq(node.get('adid'), 'someadid')
                self.eq(node.get('aaid'), 'somestr')
                self.eq(node.get('idfa'), 'someotherstr')
                self.eq(node.get('name'), 'robert grey')
                self.eq(node.get('email'), 'clown@vertex.link')
                self.eq(node.get('acct'), ('vertex.link', 'clown'))
                self.eq(node.get('app'), softguid)
                self.eq(node.get('data'), {'some key': 'some valu', 'BEEP': 1})

    async def test_telco_imei(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                # proper value
                node = await snap.addNode('tel:mob:imei', '490154203237518')
                self.eq(node.ndef[1], 490154203237518)
                self.eq(node.get('serial'), 323751)
                self.eq(node.get('tac'), 49015420)
                # One without the check bit (it gets added)
                node = await snap.addNode('tel:mob:imei', '39015420323751')
                self.eq(node.ndef[1], 390154203237519)
                # Invalid checksum
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('tel:mob:imei', 490154203237519))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('tel:mob:imei', '20'))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('tel:mob:imei', 'hehe'))

    async def test_telco_imsi(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                node = await snap.addNode('tel:mob:imsi', '310150123456789')
                self.eq(node.ndef[1], 310150123456789)
                self.eq(node.get('mcc'), '310')
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('tel:mob:imsi', 'hehe'))
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('tel:mob:imsi', 1111111111111111))

    async def test_telco_phone(self):
        async with self.getTestCore() as core:
            t = core.model.type('tel:phone')
            norm, subs = t.norm('123 456 7890')
            self.eq(norm, '1234567890')
            self.eq(subs, {'subs': {'loc': 'us'}})

            norm, subs = t.norm('123 456 \udcfe7890')
            self.eq(norm, '1234567890')

            norm, subs = t.norm(1234567890)
            self.eq(norm, '1234567890')
            self.eq(subs, {'subs': {'loc': 'us'}})

            norm, subs = t.norm('+1911')
            self.eq(norm, '1911')
            self.eq(subs, {'subs': {'loc': 'us'}})

            self.eq(t.repr('12345678901'), '+1 (234) 567-8901')
            self.eq(t.repr('9999999999'), '+9999999999')

            self.raises(s_exc.BadTypeValu, t.norm, -1)
            self.raises(s_exc.BadTypeValu, t.norm, '+()*')

            async with await core.snap() as snap:
                node = await snap.addNode('tel:phone', '+1 (703) 555-1212')
                self.eq(node.ndef[1], '17035551212')
                self.eq(node.get('loc'), 'us')
                node = await snap.addNode('tel:phone', '+1 (703) 555-2424')
                # Esnap search
                nodes = await snap.nodes('tel:phone=17035552424')
                self.len(1, nodes)
                self.eq(nodes[0].ndef[1], '17035552424')
                # Prefix search
                nodes = await snap.nodes('tel:phone=1703555*')
                self.len(2, nodes)

    async def test_telco_call(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                guid = s_common.guid()
                props = {
                    'src': '+1 (703) 555-1212',
                    'dst': '123 456 7890',
                    'time': '2001',
                    'duration': 90,
                    'connected': True,
                    'text': 'I said some stuff',
                    'file': 'sha256:' + 64 * 'f',
                }
                node = await snap.addNode('tel:call', guid, props)
                self.eq(node.ndef[1], guid)
                self.eq(node.get('src'), '17035551212')
                self.eq(node.get('dst'), '1234567890')
                self.eq(node.get('time'), 978307200000)
                self.eq(node.get('duration'), 90)
                self.eq(node.get('connected'), True)
                self.eq(node.get('text'), 'I said some stuff')
                self.eq(node.get('file'), 'sha256:' + 64 * 'f')

    async def test_telco_txtmesg(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                guid = s_common.guid()
                props = {
                    'from': '+1 (703) 555-1212',
                    'to': '123 456 7890',
                    'recipients': ('567 890 1234', '555 444 3333'),
                    'svctype': 'sms',
                    'time': '2001',
                    'text': 'I wrote some stuff',
                    'file': 'sha256:' + 64 * 'b',
                }
                node = await snap.addNode('tel:txtmesg', guid, props)
                self.eq(node.ndef[1], guid)
                self.eq(node.get('from'), '17035551212')
                self.eq(node.get('to'), '1234567890')
                self.eq(node.get('recipients'), ('5554443333', '5678901234'))
                self.eq(node.get('svctype'), 'sms')
                self.eq(node.get('time'), 978307200000)
                self.eq(node.get('text'), 'I wrote some stuff')
                self.eq(node.get('file'), 'sha256:' + 64 * 'b')

                # add other valid message types
                guid = s_common.guid()
                node = await snap.addNode('tel:txtmesg', guid, {'svctype': 'mms'})
                self.eq(node.ndef[1], guid)

                guid = s_common.guid()
                node = await snap.addNode('tel:txtmesg', guid, {'svctype': ' MMS'})
                self.eq(node.ndef[1], guid)

                guid = s_common.guid()
                node = await snap.addNode('tel:txtmesg', guid, {'svctype': 'rcs'})
                self.eq(node.ndef[1], guid)

                # no message type specified
                guid = s_common.guid()
                node = await snap.addNode('tel:txtmesg', guid, {'text': 'no message type'})
                self.eq(node.ndef[1], guid)

                # add bad svc type
                guid = s_common.guid()
                await self.asyncraises(s_exc.BadTypeValu, snap.addNode('tel:txtmesg', guid, {'svctype': 'foo'}))



import time
import types
import calendar
import datetime
import functools

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.cli as s_cli
import synapse.lib.cmd as s_cmd
import synapse.lib.time as s_time
import synapse.lib.parser as s_parser

StatHelp = '''
Gives detailed information about a single cron job.

Syntax:
    cron stat <iden prefix>

Notes:
    Any prefix that matches exactly one valid cron job iden is accepted.
'''

DelHelp = '''
Deletes a single cron job.

Syntax:
    cron del|rm <iden prefix>

Notes:
    Any prefix that matches exactly one valid cron job iden is accepted.
'''

ListHelp = '''
List existing cron jobs in a cortex.

Syntax:
    cron list|ls

Example:
    user       iden       en? rpt? now? err? # start last start       last end         query
    root       029ce7bd.. Y   Y    N           17863 2019-06-11T21:47 2019-06-11T21:47 exec foo
    root       06b46533.. Y   Y    N           18140 2019-06-11T21:48 2019-06-11T21:48 exec bar
'''

ModHelp = '''
Changes an existing cron job's query.

Syntax:
    cron mod|edit <iden prefix> <new query>

Notes:
    Any prefix that matches exactly one valid cron iden is accepted.
'''

EnableHelp = '''
Enable an existing cron job.

Syntax:
    cron enable <iden prefix>

Notes:
    Any prefix that matches exactly one valid cron iden is accepted.
'''

DisableHelp = '''
Disable an existing cron job.

Syntax:
    cron disable <iden prefix>

Notes:
    Any prefix that matches exactly one valid cron iden is accepted.
'''

AddHelp = '''
Add a recurring cron job to a cortex.

Syntax:
    cron add [optional arguments] {query}

    --minute, -M int[,int...][=]
    --hour, -H
    --day, -d
    --month, -m
    --year, -y

       *or:*

    [--hourly <min> |
     --daily <hour>:<min> |
     --monthly <day>:<hour>:<min> |
     --yearly <month>:<day>:<hour>:<min>]

Notes:
    All times are interpreted as UTC.

    All arguments are interpreted as the job period, unless the value ends in
    an equals sign, in which case the argument is interpreted as the recurrence
    period.  Only one recurrence period parameter may be specified.

    Currently, a fixed unit must not be larger than a specified recurrence
    period.  i.e. '--hour 7 --minute +15' (every 15 minutes from 7-8am?) is not
    supported.

    Value values for fixed hours are 0-23 on a 24-hour clock where midnight is 0.

    If the --day parameter value does not start with in '+' and is an integer, it is
    interpreted as a fixed day of the month.  A negative integer may be
    specified to count from the end of the month with -1 meaning the last day
    of the month.  All fixed day values are clamped to valid days, so for
    example '-d 31' will run on February 28.

    If the fixed day parameter is a value in ([Mon, Tue, Wed, Thu, Fri, Sat,
    Sun] if locale is set to English) it is interpreted as a fixed day of the
    week.

    Otherwise, if the parameter value starts with a '+', then it is interpreted
    as an recurrence interval of that many days.

    If no plus-sign-starting parameter is specified, the recurrence period
    defaults to the unit larger than all the fixed parameters.   e.g. '-M 5'
    means every hour at 5 minutes past, and -H 3, -M 1 means 3:01 every day.

    At least one optional parameter must be provided.

    All parameters accept multiple comma-separated values.  If multiple
    parameters have multiple values, all combinations of those values are used.

    All fixed units not specified lower than the recurrence period default to
    the lowest valid value, e.g. -m +2 will be scheduled at 12:00am the first of
    every other month.  One exception is the largest fixed value is day of the
    week, then the default period is set to be a week.

    A month period with a day of week fixed value is not currently supported.

    Fixed-value year (i.e. --year 2019) is not supported.  See the 'at'
    command for one-time cron jobs.

    As an alternative to the above options, one may use exactly one of
    --hourly, --daily, --monthly, --yearly with a colon-separated list of
    fixed parameters for the value.  It is an error to use both the individual
    options and these aliases at the same time.

Examples:
    Run a query every last day of the month at 3 am
    cron add -H 3 -d-1 {#foo}

    Run a query every 8 hours
    cron add -H +8 {#foo}

    Run a query every Wednesday and Sunday at midnight and noon
    cron add -H 0,12 -d Wed,Sun {#foo}

    Run a query every other day at 3:57pm
    cron add -d +2 -M 57 -H 15 {#foo}
'''

class Cron(s_cli.Cmd):
    '''
Manages cron jobs in a cortex.

Cron jobs are rules persistently stored in a cortex such that storm queries
automatically run on a time schedule.

Cron jobs may be be recurring or one-time.  Use the 'at' command to add
one-time jobs.

A subcommand is required.  Use 'cron -h' for more detailed help.  '''
    _cmd_name = 'cron'

    _cmd_syntax = (
        ('line', {'type': 'glob'}),  # type: ignore
    )

    async def _match_idens(self, core, prefix):
        '''
        Returns the iden that starts with prefix.  Prints out error and returns None if it doesn't match
        exactly one.
        '''
        idens = [cron['iden'] for cron in await core.listCronJobs()]
        matches = [iden for iden in idens if iden.startswith(prefix)]
        if len(matches) == 1:
            return matches[0]
        elif len(matches) == 0:
            self.printf('Error: provided iden does not match any valid authorized cron job')
        else:
            self.printf('Error: provided iden matches more than one cron job')
        return None

    def _make_argparser(self):

        parser = s_cmd.Parser(prog='cron', outp=self, description=self.__doc__)

        subparsers = parser.add_subparsers(title='subcommands', required=True, dest='cmd',
                                           parser_class=functools.partial(s_cmd.Parser, outp=self))

        subparsers.add_parser('list', aliases=['ls'], help="List cron jobs you're allowed to manipulate",
                              usage=ListHelp)

        parser_add = subparsers.add_parser('add', help='add a cron job', usage=AddHelp)
        parser_add.add_argument('--minute', '-M')
        parser_add.add_argument('--hour', '-H')
        parser_add.add_argument('--day', '-d', help='day of week, day of month or number of days')
        parser_add.add_argument('--month', '-m')
        parser_add.add_argument('--year', '-y')
        group = parser_add.add_mutually_exclusive_group()
        group.add_argument('--hourly')
        group.add_argument('--daily')
        group.add_argument('--monthly')
        group.add_argument('--yearly')
        parser_add.add_argument('query', help='Storm query in curly braces')

        parser_del = subparsers.add_parser('del', aliases=['rm'], help='delete a cron job', usage=DelHelp)
        parser_del.add_argument('prefix', help='Cron job iden prefix')

        parser_stat = subparsers.add_parser('stat', help='details a cron job', usage=StatHelp)
        parser_stat.add_argument('prefix', help='Cron job iden prefix')

        parser_mod = subparsers.add_parser('mod', aliases=['edit'], help='change an existing cron job', usage=ModHelp)
        parser_mod.add_argument('prefix', help='Cron job iden prefix')
        parser_mod.add_argument('query', help='New Storm query in curly braces')

        parser_en = subparsers.add_parser('enable', help='enable an existing cron job', usage=EnableHelp)
        parser_en.add_argument('prefix', help='Cron job iden prefix')

        parser_dis = subparsers.add_parser('disable', help='disable an existing cron job', usage=DisableHelp)
        parser_dis.add_argument('prefix', help='Cron job iden prefix')

        return parser

    @staticmethod
    def _parse_weekday(val):
        ''' Try to match a day-of-week abbreviation, then try a day-of-week full name '''
        val = val.title()
        try:
            return list(calendar.day_abbr).index(val)
        except ValueError:
            try:
                return list(calendar.day_name).index(val)
            except ValueError:
                return None

    @staticmethod
    def _parse_incval(incunit, incval):
        ''' Parse a non-day increment value. Should be an integer or a comma-separated integer list. '''
        try:
            retn = [int(val) for val in incval.split(',')]
        except ValueError:
            return None

        return retn[0] if len(retn) == 1 else retn

    @staticmethod
    def _parse_req(requnit, reqval):
        ''' Parse a non-day fixed value '''
        assert reqval[0] != '='

        try:
            retn = []
            for val in reqval.split(','):
                if requnit == 'month':
                    if reqval[0].isdigit():
                        retn.append(int(reqval))  # must be a month (1-12)
                    else:
                        try:
                            retn.append(list(calendar.month_abbr).index(val.title()))
                        except ValueError:
                            retn.append(list(calendar.month_name).index(val.title()))
                else:
                    retn.append(int(val))
        except ValueError:
            return None

        if not retn:
            return None

        return retn[0] if len(retn) == 1 else retn

    @staticmethod
    def _parse_day(optval):
        ''' Parse a --day argument '''
        isreq = not optval.startswith('+')
        if not isreq:
            optval = optval[1:]

        try:
            retnval = []
            unit = None
            for val in optval.split(','):
                if not val:
                    raise ValueError
                if val[-1].isdigit():
                    newunit = 'dayofmonth' if isreq else 'day'
                    if unit is None:
                        unit = newunit
                    elif newunit != unit:
                        raise ValueError
                    retnval.append(int(val))
                else:
                    newunit = 'dayofweek'
                    if unit is None:
                        unit = newunit
                    elif newunit != unit:
                        raise ValueError

                    weekday = Cron._parse_weekday(val)
                    if weekday is None:
                        raise ValueError
                    retnval.append(weekday)
            if len(retnval) == 0:
                raise ValueError
        except ValueError:
            return None, None
        if len(retnval) == 1:
            retnval = retnval[0]
        return unit, retnval

    def _parse_alias(self, opts):
        retn = types.SimpleNamespace()
        retn.query = opts.query

        if opts.hourly is not None:
            retn.hour = '+1'
            retn.minute = str(int(opts.hourly))
            return retn

        if opts.daily is not None:
            fields = time.strptime(opts.daily, '%H:%M')
            retn.day = '+1'
            retn.hour = str(fields.tm_hour)
            retn.minute = str(fields.tm_min)
            return retn

        if opts.monthly is not None:
            day, rest = opts.monthly.split(':', 1)
            fields = time.strptime(rest, '%H:%M')
            retn.month = '+1'
            retn.day = day
            retn.hour = str(fields.tm_hour)
            retn.minute = str(fields.tm_min)
            return retn

        if opts.yearly is not None:
            fields = opts.yearly.split(':')
            if len(fields) != 4:
                raise ValueError(f'Failed to parse parameter {opts.yearly}')
            retn.year = '+1'
            retn.month, retn.day, retn.hour, retn.minute = fields
            return retn

        return None

    async def _handle_add(self, core, opts):
        incunit = None
        incval = None
        reqdict = {}
        valinfo = {  # unit: (minval, next largest unit)
            'month': (1, 'year'),
            'dayofmonth': (1, 'month'),
            'hour': (0, 'day'),
            'minute': (0, 'hour'),
        }

        if not opts.query.startswith('{'):
            self.printf('Error: query parameter must start with {')
            return

        try:
            alias_opts = self._parse_alias(opts)
        except ValueError as e:
            self.printf(f'Error: Failed to parse ..ly parameter: {" ".join(e.args)}')
            return

        if alias_opts:
            if opts.year or opts.month or opts.day or opts.hour or opts.minute:
                self.printf('Error: may not use both alias (..ly) and explicit options at the same time')
                return
            opts = alias_opts

        for optname in ('year', 'month', 'day', 'hour', 'minute'):
            optval = getattr(opts, optname, None)

            if optval is None:
                if incunit is None and not reqdict:
                    continue
                # The option isn't set, but a higher unit is.  Go ahead and set the required part to the lowest valid
                # value, e.g. so -m 2 would run on the *first* of every other month at midnight
                if optname == 'day':
                    reqdict['dayofmonth'] = 1
                else:
                    reqdict[optname] = valinfo[optname][0]
                continue

            isreq = not optval.startswith('+')

            if optname == 'day':
                unit, val = self._parse_day(optval)
                if val is None:
                    self.printf(f'Error: failed to parse day value "{optval}"')
                    return
                if unit == 'dayofweek':
                    if incunit is not None:
                        self.printf('Error: May not provide a recurrence value with day of week')
                        return
                    if reqdict:
                        self.printf('Error: may not fix month or year with day of week')
                        return
                    incunit, incval = unit, val
                elif unit == 'day':
                    incunit, incval = unit, val
                else:
                    assert unit == 'dayofmonth'
                    reqdict[unit] = val
                continue

            if not isreq:
                if incunit is not None:
                    self.printf('Error: may not provide more than 1 recurrence parameter')
                    return
                if reqdict:
                    self.printf('Error: fixed unit may not be larger than recurrence unit')
                    return
                incunit = optname
                incval = self._parse_incval(optname, optval)
                if incval is None:
                    self.printf('Error: failed to parse parameter')
                    return
                continue

            if optname == 'year':
                self.printf('Error: year may not be a fixed value')
                return

            reqval = self._parse_req(optname, optval)
            if reqval is None:
                self.printf(f'Error: failed to parse fixed parameter "{optval}"')
                return
            reqdict[optname] = reqval

        # If not set, default (incunit, incval) to (1, the next largest unit)
        if incunit is None:
            if not reqdict:
                self.printf('Error: must provide at least one optional argument')
                return
            requnit = next(iter(reqdict))  # the first key added is the biggest unit
            incunit = valinfo[requnit][1]
            incval = 1

        # Remove the curly braces
        query = opts.query[1:-1]
        cdef = {'storm': query,
                'reqs': reqdict,
                'incunit': incunit,
                'incvals': incval,
                }
        newcdef = await core.addCronJob(cdef)
        self.printf(f'Created cron job {newcdef["iden"]}')

    @staticmethod
    def _format_timestamp(ts):
        # N.B. normally better to use fromtimestamp with UTC timezone, but we don't want timezone to print out
        return datetime.datetime.utcfromtimestamp(ts).isoformat(timespec='minutes')

    async def _handle_list(self, core, opts):
        cronlist = await core.listCronJobs()

        if not cronlist:
            self.printf('No cron jobs found')
            return
        self.printf(
            f'{"user":10} {"iden":10} {"en?":3} {"rpt?":4} {"now?":4} {"err?":4} '
            f'{"# start":7} {"last start":16} {"last end":16} {"query"}')

        for cron in cronlist:
            iden = cron.get('iden')

            idenf = iden[:8] + '..'
            user = cron.get('username') or '<None>'
            query = cron.get('query') or '<missing>'
            isrecur = 'Y' if cron.get('recur') else 'N'
            isrunning = 'Y' if cron.get('isrunning') else 'N'
            enabled = 'Y' if cron.get('enabled') else 'N'
            startcount = cron.get('startcount') or 0
            laststart = cron.get('laststarttime')
            laststart = 'Never' if laststart is None else self._format_timestamp(laststart)
            lastend = cron.get('lastfinishtime')
            lastend = 'Never' if lastend is None else self._format_timestamp(lastend)
            result = cron.get('lastresult')
            iserr = 'X' if result is not None and not result.startswith('finished successfully') else ' '

            self.printf(
                f'{user:10} {idenf:10} {enabled:3} {isrecur:4} {isrunning:4} {iserr:4} '
                f'{startcount:7} {laststart:16} {lastend:16} {query}')

    async def _handle_mod(self, core, opts):
        prefix = opts.prefix
        query = opts.query
        if not query.startswith('{'):
            self.printf('Error:  expected second argument to start with {')
            return
        # remove the curly braces
        query = query[1:-1]
        iden = await self._match_idens(core, prefix)
        if iden is None:
            return
        await core.updateCronJob(iden, query)
        self.printf(f'Modified cron job {iden}')

    async def _handle_enable(self, core, opts):
        prefix = opts.prefix
        iden = await self._match_idens(core, prefix)
        if iden is None:
            return
        await core.enableCronJob(iden)
        self.printf(f'Enabled cron job {iden}')

    async def _handle_disable(self, core, opts):
        prefix = opts.prefix
        iden = await self._match_idens(core, prefix)
        if iden is None:
            return
        await core.disableCronJob(iden)
        self.printf(f'Disabled cron job {iden}')

    async def _handle_del(self, core, opts):
        prefix = opts.prefix
        iden = await self._match_idens(core, prefix)
        if iden is None:
            return
        await core.delCronJob(iden)
        self.printf(f'Deleted cron job {iden}')

    async def _handle_stat(self, core, opts):
        ''' Prints details about a particular cron job. Not actually a different API call '''
        prefix = opts.prefix
        crons = await core.listCronJobs()
        idens = [cron['iden'] for cron in crons]
        matches = [iden for iden in idens if iden.startswith(prefix)]
        if len(matches) == 0:
            self.printf('Error: provided iden does not match any valid authorized cron job')
            return
        elif len(matches) > 1:
            self.printf('Error: provided iden matches more than one cron job')
            return

        iden = matches[0]
        cron = [cron for cron in crons if cron.get('iden') == iden][0]

        user = cron.get('username') or '<None>'
        query = cron.get('query') or '<missing>'
        isrecur = 'Yes' if cron.get('recur') else 'No'
        enabled = 'Yes' if cron.get('enabled') else 'No'
        startcount = cron.get('startcount') or 0
        recs = cron.get('recs', [])
        laststart = cron.get('laststarttime')
        lastend = cron.get('lastfinishtime')
        laststart = 'Never' if laststart is None else self._format_timestamp(laststart)
        lastend = 'Never' if lastend is None else self._format_timestamp(lastend)
        lastresult = cron.get('lastresult') or '<None>'

        self.printf(f'iden:            {iden}')
        self.printf(f'user:            {user}')
        self.printf(f'enabled:         {enabled}')
        self.printf(f'recurring:       {isrecur}')
        self.printf(f'# starts:        {startcount}')
        self.printf(f'last start time: {laststart}')
        self.printf(f'last end time:   {lastend}')
        self.printf(f'last result:     {lastresult}')
        self.printf(f'query:           {query}')
        if not recs:
            self.printf(f'entries:         <None>')
        else:
            self.printf(f'entries:         {"incunit":10} {"incval":6} {"required"}')
            for reqdict, incunit, incval in recs:
                reqdict = reqdict or '<None>'
                incunit = '<None>' if incunit is None else incunit
                incval = '<None>' if incval is None else incval
                self.printf(f'                 {incunit:10} {incval:6} {reqdict}')

    async def runCmdOpts(self, opts):

        s_common.deprecated('cmdr> cron')

        line = opts.get('line')
        if line is None:
            self.printf(self.__doc__)
            return

        core = self.getCmdItem()

        argv = s_parser.Parser(line).cmdrargs()
        try:
            opts = self._make_argparser().parse_args(argv)
        except s_exc.ParserExit:
            return

        handlers = {
            'add': self._handle_add,
            'del': self._handle_del,
            'rm': self._handle_del,
            'disable': self._handle_disable,
            'enable': self._handle_enable,
            'list': self._handle_list,
            'ls': self._handle_list,
            'mod': self._handle_mod,
            'edit': self._handle_mod,
            'stat': self._handle_stat,
        }
        await handlers[opts.cmd](core, opts)

class At(s_cli.Cmd):
    '''
Adds a non-recurring cron job.

It will execute a Storm query at one or more specified times.

List/details/deleting cron jobs created with 'at' use the same commands as
other cron jobs:  cron list/stat/del respectively.

Syntax:
    at (time|+time delta)+ {query}

Notes:
    This command accepts one or more time specifications followed by exactly
    one storm query in curly braces.  Each time specification may be in synapse
    time delta format (e.g + 1 day) or synapse time format (e.g.
    20501217030432101).  Seconds will be ignored, as cron jobs' granularity is
    limited to minutes.

    All times are interpreted as UTC.

    The other option for time specification is a relative time from now.  This
    consists of a plus sign, a positive integer, then one of 'minutes, hours,
    days'.

    Note that the record for a cron job is stored until explicitly deleted via
    "cron del".

Examples:
    # Run a storm query in 5 minutes
    at +5 minutes {[inet:ipv4=1]}

    # Run a storm query tomorrow and in a week
    at +1 day +7 days {[inet:ipv4=1]}

    # Run a query at the end of the year Zulu
    at 20181231Z2359 {[inet:ipv4=1]}
'''
    _cmd_name = 'at'

    _cmd_syntax = (
        ('line', {'type': 'glob'}),  # type: ignore
    )

    def _make_argparser(self):
        parser = s_cmd.Parser(prog='at', outp=self, description=self.__doc__)
        parser.add_argument('args', nargs='+', help='date | delta| {query})')
        return parser

    async def runCmdOpts(self, opts):

        s_common.deprecated('cmdr> at')

        line = opts.get('line')
        if line is None:
            self.printf(self.__doc__)
            return

        core = self.getCmdItem()

        argv = s_parser.Parser(line).cmdrargs()
        # Currently, using an argparser is overkill for this command.  Using for future extensibility (and help).
        try:
            opts = self._make_argparser().parse_args(argv)
        except s_exc.ParserExit:
            return

        query = None
        consumed_next = False

        tslist = []
        # TODO: retrieve time from cortex in case of wrong cmdr time
        now = time.time()

        for pos, arg in enumerate(opts.args):
            try:
                if consumed_next:
                    consumed_next = False
                    continue

                if arg.startswith('{'):
                    if query is not None:
                        self.printf('Error: only a single query is allowed')
                        return
                    query = arg[1:-1]
                    continue

                if arg.startswith('+'):
                    if arg[-1].isdigit():
                        if pos == len(opts.args) - 1:
                            self.printf('Time delta missing unit')
                            return
                        arg = f'{arg} {opts.args[pos + 1]}'
                        consumed_next = True
                    ts = now + s_time.delta(arg) / 1000.0
                    tslist.append(ts)
                    continue

                ts = s_time.parse(arg) / 1000.0
                tslist.append(ts)
            except (ValueError, s_exc.BadTypeValu):
                self.printf(f'Error: Trouble parsing "{arg}"')
                return

        if query is None:
            self.printf('Error: Missing query argument')
            return

        def _ts_to_reqdict(ts):
            dt = datetime.datetime.fromtimestamp(ts, datetime.timezone.utc)
            return {
                'minute': dt.minute,
                'hour': dt.hour,
                'dayofmonth': dt.day,
                'month': dt.month,
                'year': dt.year
            }

        if not tslist:
            self.printf('Error: at least one requirement must be provided')
            return

        reqdicts = [_ts_to_reqdict(ts) for ts in tslist]

        cdef = {'storm': query,
                'reqs': reqdicts,
                'incunit': None,
                'incvals': None,
                }

        newcdef = await core.addCronJob(cdef)
        self.printf(f'Created cron job {newcdef["iden"]}')



import logging

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.stormtypes as s_stormtypes


logger = logging.getLogger(__name__)

stormlogger = logging.getLogger('synapse.storm.log')

@s_stormtypes.registry.registerLib
class LoggerLib(s_stormtypes.Lib):
    '''
    A Storm library which implements server side logging. These messages are logged to the
    ``synapse.storm.log`` logger.
    '''
    _storm_locals = (
        {'name': 'debug', 'desc': '''
        Log a message to the Cortex at the debug log level.

        Notes:
            This requires the ``storm.lib.log.debug`` permission to use.

        Examples:
            Log a debug message::

                $lib.log.debug('I am a debug message!')

            Log a debug message with extra information::

                $lib.log.debug('Extra information included here.', extra=({"key": $valu}))''',
         'type': {'type': 'function', '_funcname': '_logDebug',
                  'args': (
                      {'name': 'mesg', 'type': 'str', 'desc': 'The message to log.'},
                      {'name': 'extra', 'type': 'dict', 'desc': 'Extra key / value pairs to include when structured '
                                                                'logging is enabled on the Cortex.',
                       'default': None, },
                  ),
                  'returns': {'type': 'null'},
                  }
         },
        {'name': 'info', 'desc': '''
            Log a message to the Cortex at the info log level.

            Notes:
                This requires the ``storm.lib.log.info`` permission to use.

            Examples:
                Log an info message::

                    $lib.log.info('I am a info message!')

                Log an info message with extra information::

                    $lib.log.info('Extra information included here.', extra=({"key": $valu}))''',
         'type': {'type': 'function', '_funcname': '_logInfo',
                  'args': (
                      {'name': 'mesg', 'type': 'str', 'desc': 'The message to log.'},
                      {'name': 'extra', 'type': 'dict', 'desc': 'Extra key / value pairs to include when structured '
                                                                'logging is enabled on the Cortex.',
                       'default': None, },
                  ),
                  'returns': {'type': 'null'},
                  }
         },
        {'name': 'warning', 'desc': '''
            Log a message to the Cortex at the warning log level.

            Notes:
                This requires the ``storm.lib.log.warning`` permission to use.

            Examples:
                Log a warning message::

                    $lib.log.warning('I am a warning message!')

                Log a warning message with extra information::

                    $lib.log.warning('Extra information included here.', extra=({"key": $valu}))''',
         'type': {'type': 'function', '_funcname': '_logDebug',
                  'args': (
                      {'name': 'mesg', 'type': 'str', 'desc': 'The message to log.'},
                      {'name': 'extra', 'type': 'dict', 'desc': 'Extra key / value pairs to include when structured '
                                                                'logging is enabled on the Cortex.',
                       'default': None, },
                  ),
                  'returns': {'type': 'null'},
                  }
         },
        {'name': 'error', 'desc': '''
            Log a message to the Cortex at the error log level.

            Notes:
                This requires the ``storm.lib.log.error`` permission to use.

            Examples:
                Log an error message::

                    $lib.log.error('I am a error message!')

                Log an error message with extra information::

                    $lib.log.error('Extra information included here.', extra=({"key": $valu}))''',
         'type': {'type': 'function', '_funcname': '_logDebug',
                  'args': (
                      {'name': 'mesg', 'type': 'str', 'desc': 'The message to log.'},
                      {'name': 'extra', 'type': 'dict', 'desc': 'Extra key / value pairs to include when structured '
                                                                'logging is enabled on the Cortex.',
                       'default': None, },
                  ),
                  'returns': {'type': 'null'},
                  }
         },
    )

    _storm_lib_path = ('log',)

    def getObjLocals(self):
        return {
            'info': self._logInfo,
            'debug': self._logDebug,
            'error': self._logError,
            'warning': self._logWarning,
        }

    async def _getExtra(self, extra=None):
        if extra is None:
            return extra
        extra = await s_stormtypes.toprim(extra)
        if extra and not isinstance(extra, dict):
            mesg = f'extra provided to log call must be a dictionary compatible type. Got {extra.__class__.__name__} ' \
                   f'instead.'
            raise s_exc.BadArg(mesg=mesg, arg='extra')
        extra = {'synapse': extra}
        return extra

    @s_stormtypes.stormfunc(readonly=True)
    async def _logDebug(self, mesg, extra=None):
        self.runt.confirm(('storm', 'lib', 'log', 'debug'))
        mesg = await s_stormtypes.tostr(mesg)
        extra = await self._getExtra(extra)
        stormlogger.debug(mesg, extra=extra)
        return None

    @s_stormtypes.stormfunc(readonly=True)
    async def _logInfo(self, mesg, extra=None):
        self.runt.confirm(('storm', 'lib', 'log', 'info'))
        mesg = await s_stormtypes.tostr(mesg)
        extra = await self._getExtra(extra)
        stormlogger.info(mesg, extra=extra)
        return None

    @s_stormtypes.stormfunc(readonly=True)
    async def _logWarning(self, mesg, extra=None):
        self.runt.confirm(('storm', 'lib', 'log', 'warning'))
        mesg = await s_stormtypes.tostr(mesg)
        extra = await self._getExtra(extra)
        stormlogger.warning(mesg, extra=extra)
        return None

    @s_stormtypes.stormfunc(readonly=True)
    async def _logError(self, mesg, extra=None):
        self.runt.confirm(('storm', 'lib', 'log', 'error'))
        mesg = await s_stormtypes.tostr(mesg)
        extra = await self._getExtra(extra)
        stormlogger.error(mesg, extra=extra)
        return None
